@@ -1,4 +1,5 @@
 /*
+ * Copyright 2008-2011 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -12,9 +13,7 @@
  * General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "command_mode.h"
@@ -28,6 +27,7 @@
 #include "browser.h"
 #include "filters.h"
 #include "player.h"
+#include "output.h"
 #include "editable.h"
 #include "lib.h"
 #include "pl.h"
@@ -39,21 +39,22 @@
 #include "xstrjoin.h"
 #include "misc.h"
 #include "path.h"
-#include "format_print.h"
 #include "spawn.h"
 #include "utils.h"
 #include "list.h"
 #include "debug.h"
 #include "load_dir.h"
+#ifdef HAVE_CONFIG
 #include "config/datadir.h"
+#endif
 #include "help.h"
+#include "op.h"
 
 #include <stdlib.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <dirent.h>
-#include <pwd.h>
 
 #if defined(__sun__)
 #include <ncurses.h>
@@ -67,63 +68,6 @@ static char *history_search_text = NULL;
 static int arg_expand_cmd = -1;
 static int prev_view = -1;
 
-static char *get_home_dir(const char *username)
-{
-	struct passwd *passwd;
-
-	if (username == NULL)
-		return xstrdup(home_dir);
-	passwd = getpwnam(username);
-	if (passwd == NULL)
-		return NULL;
-	/* don't free passwd */
-	return xstrdup(passwd->pw_dir);
-}
-
-static char *expand_filename(const char *name)
-{
-	if (name[0] == '~') {
-		char *slash;
-
-		slash = strchr(name, '/');
-		if (slash) {
-			char *username, *home;
-
-			if (slash - name - 1 > 0) {
-				/* ~user/... */
-				username = xstrndup(name + 1, slash - name - 1);
-			} else {
-				/* ~/... */
-				username = NULL;
-			}
-			home = get_home_dir(username);
-			free(username);
-			if (home) {
-				char *expanded;
-
-				expanded = xstrjoin(home, slash);
-				free(home);
-				return expanded;
-			} else {
-				return xstrdup(name);
-			}
-		} else {
-			if (name[1] == 0) {
-				return xstrdup(home_dir);
-			} else {
-				char *home;
-
-				home = get_home_dir(name + 1);
-				if (home)
-					return home;
-				return xstrdup(name);
-			}
-		}
-	} else {
-		return xstrdup(name);
-	}
-}
-
 /* view {{{ */
 
 void view_clear(int view)
@@ -240,30 +184,43 @@ void view_load(int view, char *arg)
 	}
 }
 
-static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filenamep)
+static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filenamep, save_ti_cb save_ti)
 {
 	char *filename = *filenamep;
 
 	if (arg) {
-		free(filename);
-		filename = xstrdup(arg);
-		*filenamep = filename;
+		if (strcmp(arg, "-") == 0) {
+			filename = (char *) arg;
+		} else {
+			free(filename);
+			filename = xstrdup(arg);
+			*filenamep = filename;
+		}
+	} else if (!filename) {
+		error_msg("need a file as argument, no default stored yet");
+		return;
 	}
 
 	editable_lock();
-	if (cmus_save(for_each_ti, filename) == -1)
+	if (save_ti(for_each_ti, filename) == -1)
 		error_msg("saving '%s': %s", filename, strerror(errno));
 	editable_unlock();
 }
 
-void view_save(int view, char *arg)
+void view_save(int view, char *arg, int to_stdout, int filtered, int extended)
 {
-	if (arg) {
-		char *tmp;
+	char **dest;
+	save_ti_cb     save_ti         = extended ? cmus_save_ext         : cmus_save;
+	for_each_ti_cb lib_for_each_ti = filtered ? lib_for_each_filtered : lib_for_each;
 
-		tmp = expand_filename(arg);
-		arg = path_absolute(tmp);
-		free(tmp);
+	if (arg) {
+		if (to_stdout) {
+			arg = xstrdup(arg);
+		} else {
+			char *tmp = expand_filename(arg);
+			arg = path_absolute(tmp);
+			free(tmp);
+		}
 	}
 
 	switch (view) {
@@ -271,12 +228,20 @@ void view_save(int view, char *arg)
 	case SORTED_VIEW:
 		if (worker_has_job(JOB_TYPE_LIB))
 			goto worker_running;
-		do_save(lib_for_each, arg, &lib_filename);
+		dest = extended ? &lib_ext_filename : &lib_filename;
+		do_save(lib_for_each_ti, arg, dest, save_ti);
 		break;
 	case PLAYLIST_VIEW:
 		if (worker_has_job(JOB_TYPE_PL))
 			goto worker_running;
-		do_save(pl_for_each, arg, &pl_filename);
+		dest = extended ? &pl_ext_filename : &pl_filename;
+		do_save(pl_for_each, arg, dest, save_ti);
+		break;
+	case QUEUE_VIEW:
+		if (worker_has_job(JOB_TYPE_QUEUE))
+			goto worker_running;
+		dest = extended ? &play_queue_ext_filename : &play_queue_filename;
+		do_save(play_queue_for_each, arg, dest, save_ti);
 		break;
 	default:
 		info_msg(":save only works in views 1 & 2 (library) and 3 (playlist)");
@@ -290,8 +255,8 @@ worker_running:
 
 /* }}} */
 
-/* only returns the last flag which is enough for it's callers */
-static int parse_flags(const char **strp, const char *flags)
+/* if only_last != 0, only return the last flag */
+static int do_parse_flags(const char **strp, const char *flags, int only_last)
 {
 	const char *str = *strp;
 	int flag = 0;
@@ -299,7 +264,7 @@ static int parse_flags(const char **strp, const char *flags)
 	if (str == NULL)
 		return flag;
 
-	while (*str) {
+	while (*str && (only_last || !flag)) {
 		if (*str != '-')
 			break;
 
@@ -336,10 +301,47 @@ static int parse_flags(const char **strp, const char *flags)
 	return flag;
 }
 
+static int parse_flags(const char **strp, const char *flags)
+{
+	return do_parse_flags(strp, flags, 1);
+}
+
+static int parse_one_flag(const char **strp, const char *flags)
+{
+	return do_parse_flags(strp, flags, 0);
+}
+
+/* is str == "...-", but not "...-- -" ? copied from do_parse_flags() */
+static int is_stdout_filename(const char *str)
+{
+	if (!str)
+		return 0;
+
+	while (*str) {
+		if (*str != '-')
+			return 0;
+		// "-"
+		if (str[1] == 0)
+			return 1;
+		// "--" or "-- "
+		if (str[1] == '-' && (str[2] == 0 || str[2] == ' '))
+			return 0;
+		// not "-?" or "-? "
+		if (str[2] && str[2] != ' ')
+			return 0;
+		str += 2;
+		while (*str == ' ')
+			str++;
+	}
+
+	return 0;
+}
+
 static int flag_to_view(int flag)
 {
 	switch (flag) {
 	case 'l':
+	case 'L':
 		return TREE_VIEW;
 	case 'p':
 		return PLAYLIST_VIEW;
@@ -392,11 +394,20 @@ static void cmd_load(char *arg)
 
 static void cmd_save(char *arg)
 {
-	int flag = parse_flags((const char **)&arg, "lp");
+	int to_stdout = is_stdout_filename(arg);
+	int flag = 0, f, extended = 0;
+
+	do {
+		f = parse_one_flag((const char **)&arg, "eLlpq");
+		if (f == 'e')
+			extended = 1;
+		else if (f)
+			flag = f;
+	} while (f > 0);
 
 	if (flag == -1)
 		return;
-	view_save(flag_to_view(flag), arg);
+	view_save(flag_to_view(flag), arg, to_stdout, flag == 'L', extended);
 }
 
 static void cmd_set(char *arg)
@@ -495,7 +506,7 @@ inside:
 		goto err;
 
 	if (count == 1) {
-		switch (tolower(*arg)) {
+		switch (tolower((unsigned char)*arg)) {
 		case 'h':
 			seek *= 60;
 		case 'm':
@@ -507,7 +518,7 @@ inside:
 	}
 
 	if (!*arg) {
-		player_seek(seek, relative);
+		player_seek(seek, relative, 1);
 		return;
 	}
 err:
@@ -521,6 +532,13 @@ static void cmd_factivate(char *arg)
 	editable_unlock();
 }
 
+static void cmd_live_filter(char *arg)
+{
+	editable_lock();
+	filters_set_live(arg);
+	editable_unlock();
+}
+
 static void cmd_filter(char *arg)
 {
 	editable_lock();
@@ -533,6 +551,11 @@ static void cmd_fset(char *arg)
 	filters_set_filter(arg);
 }
 
+static void cmd_help(char *arg)
+{
+	info_msg("To get started with cmus, read cmus-tutorial(7) and cmus(1) man pages");
+}
+
 static void cmd_invert(char *arg)
 {
 	editable_lock();
@@ -590,6 +613,12 @@ static void cmd_unmark(char *arg)
 	editable_unlock();
 }
 
+static void cmd_update_cache(char *arg)
+{
+	int flag = parse_flags((const char **)&arg, "f");
+	cmus_update_cache(flag == 'f');
+}
+
 static void cmd_cd(char *arg)
 {
 	if (arg) {
@@ -697,8 +726,14 @@ err:
 
 static void cmd_quit(char *arg)
 {
-	if (!worker_has_job(JOB_TYPE_ANY) || yes_no_query("Tracks are being added. Quit and truncate playlist(s)? [y/N]"))
-		cmus_running = 0;
+	int flag = parse_flags((const char **)&arg, "i");
+	if (!worker_has_job(JOB_TYPE_ANY)) {
+		if (flag != 'i' || yes_no_query("Quit cmus? [y/N]"))
+			cmus_running = 0;
+	} else {
+		if (yes_no_query("Tracks are being added. Quit and truncate playlist(s)? [y/N]"))
+			cmus_running = 0;
+	}
 }
 
 static void cmd_reshuffle(char *arg)
@@ -914,18 +949,21 @@ error:
 	return NULL;
 }
 
-static struct track_info **sel_tis;
-static int sel_tis_alloc;
-static int sel_tis_nr;
+struct track_info_selection {
+	struct track_info **tis;
+	int tis_alloc;
+	int tis_nr;
+};
 
 static int add_ti(void *data, struct track_info *ti)
 {
-	if (sel_tis_nr == sel_tis_alloc) {
-		sel_tis_alloc = sel_tis_alloc ? sel_tis_alloc * 2 : 8;
-		sel_tis = xrenew(struct track_info *, sel_tis, sel_tis_alloc);
+	struct track_info_selection *sel = data;
+	if (sel->tis_nr == sel->tis_alloc) {
+		sel->tis_alloc = sel->tis_alloc ? sel->tis_alloc * 2 : 8;
+		sel->tis = xrenew(struct track_info *, sel->tis, sel->tis_alloc);
 	}
 	track_info_ref(ti);
-	sel_tis[sel_tis_nr++] = ti;
+	sel->tis[sel->tis_nr++] = ti;
 	return 0;
 }
 
@@ -933,6 +971,7 @@ static void cmd_run(char *arg)
 {
 	char **av, **argv;
 	int ac, argc, i, run, files_idx = -1;
+	struct track_info_selection sel = { .tis = NULL };
 
 	if (cur_view > QUEUE_VIEW) {
 		info_msg("Command execution is supported only in views 1-4");
@@ -945,48 +984,44 @@ static void cmd_run(char *arg)
 	}
 
 	/* collect selected files (struct track_info) */
-	sel_tis = NULL;
-	sel_tis_alloc = 0;
-	sel_tis_nr = 0;
-
 	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
 		__tree_for_each_sel(add_ti, NULL, 0);
 		break;
 	case SORTED_VIEW:
-		__editable_for_each_sel(&lib_editable, add_ti, NULL, 0);
+		__editable_for_each_sel(&lib_editable, add_ti, &sel, 0);
 		break;
 	case PLAYLIST_VIEW:
-		__editable_for_each_sel(&pl_editable, add_ti, NULL, 0);
+		__editable_for_each_sel(&pl_editable, add_ti, &sel, 0);
 		break;
 	case QUEUE_VIEW:
-		__editable_for_each_sel(&pq_editable, add_ti, NULL, 0);
+		__editable_for_each_sel(&pq_editable, add_ti, &sel, 0);
 		break;
 	}
 	editable_unlock();
 
-	if (sel_tis_nr == 0) {
+	if (sel.tis_nr == 0) {
 		/* no files selected, do nothing */
 		free_str_array(av);
 		return;
 	}
-	sel_tis[sel_tis_nr] = NULL;
+	sel.tis[sel.tis_nr] = NULL;
 
 	/* build argv */
-	argv = xnew(char *, ac + sel_tis_nr + 1);
+	argv = xnew(char *, ac + sel.tis_nr + 1);
 	argc = 0;
 	if (files_idx == -1) {
 		/* add selected files after rest of the args */
 		for (i = 0; i < ac; i++)
 			argv[argc++] = av[i];
-		for (i = 0; i < sel_tis_nr; i++)
-			argv[argc++] = sel_tis[i]->filename;
+		for (i = 0; i < sel.tis_nr; i++)
+			argv[argc++] = sel.tis[i]->filename;
 	} else {
 		for (i = 0; i < files_idx; i++)
 			argv[argc++] = av[i];
-		for (i = 0; i < sel_tis_nr; i++)
-			argv[argc++] = sel_tis[i]->filename;
+		for (i = 0; i < sel.tis_nr; i++)
+			argv[argc++] = sel.tis[i]->filename;
 		for (i = files_idx; i < ac; i++)
 			argv[argc++] = av[i];
 	}
@@ -996,8 +1031,8 @@ static void cmd_run(char *arg)
 		d_print("ARG: '%s'\n", argv[i]);
 
 	run = 1;
-	if (confirm_run && (sel_tis_nr > 1 || strcmp(argv[0], "rm") == 0)) {
-		if (!yes_no_query("Execute %s for the %d selected files? [y/N]", arg, sel_tis_nr)) {
+	if (confirm_run && (sel.tis_nr > 1 || strcmp(argv[0], "rm") == 0)) {
+		if (!yes_no_query("Execute %s for the %d selected files? [y/N]", arg, sel.tis_nr)) {
 			info_msg("Aborted");
 			run = 0;
 		}
@@ -1005,7 +1040,7 @@ static void cmd_run(char *arg)
 	if (run) {
 		int status;
 
-		if (spawn(argv, &status)) {
+		if (spawn(argv, &status, 1)) {
 			error_msg("executing %s: %s", argv[0], strerror(errno));
 		} else {
 			if (WIFEXITED(status)) {
@@ -1020,23 +1055,31 @@ static void cmd_run(char *arg)
 			switch (cur_view) {
 			case TREE_VIEW:
 			case SORTED_VIEW:
-				/* this must be done before sel_tis are unreffed */
+				/* this must be done before sel.tis are unreffed */
 				free_str_array(av);
 				free(argv);
 
 				/* remove non-existed files, update tags for changed files */
-				cmus_update_tis(sel_tis, sel_tis_nr);
+				cmus_update_tis(sel.tis, sel.tis_nr, 0);
 
-				/* we don't own sel_tis anymore! */
+				/* we don't own sel.tis anymore! */
 				return;
 			}
 		}
 	}
 	free_str_array(av);
 	free(argv);
-	for (i = 0; sel_tis[i]; i++)
-		track_info_unref(sel_tis[i]);
-	free(sel_tis);
+	for (i = 0; sel.tis[i]; i++)
+		track_info_unref(sel.tis[i]);
+	free(sel.tis);
+}
+
+static void cmd_shell(char *arg)
+{
+	const char * const argv[] = { "sh", "-c", arg, NULL };
+
+	if (spawn((char **) argv, NULL, 0))
+		error_msg("executing '%s': %s", arg, strerror(errno));
 }
 
 static int get_one_ti(void *data, struct track_info *ti)
@@ -1145,16 +1188,16 @@ err:
 	return -1;
 }
 
-static int calc_vol(int val, int old, unsigned int flags)
+static int calc_vol(int val, int old, int max_vol, unsigned int flags)
 {
 	if (flags & VF_RELATIVE) {
 		if (flags & VF_PERCENTAGE)
-			val = scale_from_percentage(val, volume_max);
+			val = scale_from_percentage(val, max_vol);
 		val += old;
 	} else if (flags & VF_PERCENTAGE) {
-		val = scale_from_percentage(val, volume_max);
+		val = scale_from_percentage(val, max_vol);
 	}
-	return clamp(val, 0, volume_max);
+	return clamp(val, 0, max_vol);
 }
 
 /*
@@ -1166,7 +1209,7 @@ static void cmd_vol(char *arg)
 {
 	char **values = get_words(arg);
 	unsigned int lf, rf;
-	int l, r, ol, or;
+	int l, r;
 
 	if (values[1] && values[2])
 		goto err;
@@ -1181,10 +1224,24 @@ static void cmd_vol(char *arg)
 
 	free_str_array(values);
 
-	player_get_volume(&ol, &or);
-	l = calc_vol(l, ol, lf);
-	r = calc_vol(r, or, rf);
-	player_set_volume(l, r);
+	if (soft_vol) {
+		l = calc_vol(l, soft_vol_l, 100, lf);
+		r = calc_vol(r, soft_vol_r, 100, rf);
+		player_set_soft_volume(l, r);
+	} else {
+		int rc;
+		mixer_read_volume();
+		l = calc_vol(l, volume_l, volume_max, lf);
+		r = calc_vol(r, volume_r, volume_max, rf);
+		rc = mixer_set_volume(l, r);
+		if (rc != OP_ERROR_SUCCESS) {
+			char *msg = op_get_error_msg(rc, "can't change volume");
+			error_msg("%s", msg);
+			free(msg);
+		}
+		mixer_read_volume();
+	}
+	update_statusline();
 	return;
 err:
 	free_str_array(values);
@@ -1211,6 +1268,12 @@ static void cmd_view(char *arg)
 	}
 }
 
+static void cmd_push(char *arg)
+{
+	cmdline_set_text(arg);
+	enter_command_mode();
+}
+
 static void cmd_p_next(char *arg)
 {
 	cmus_next();
@@ -1278,12 +1341,17 @@ static for_each_sel_ti_cb view_for_each_sel[4] = {
 	pq_for_each_sel
 };
 
+/* wrapper for add_ti_cb, (void *) can't store function pointers */
+struct wrapper_cb_data {
+	add_ti_cb cb;
+};
+
 /* wrapper for void lib_add_track(struct track_info *) etc. */
 static int wrapper_cb(void *data, struct track_info *ti)
 {
-	add_ti_cb add = data;
+	struct wrapper_cb_data *add = data;
 
-	add(ti);
+	add->cb(ti);
 	return 0;
 }
 
@@ -1295,6 +1363,11 @@ static void add_from_browser(add_ti_cb add, int job_type)
 		enum file_type ft;
 		char *ret;
 
+		if (ends_with(sel, "/../") || ends_with(sel, "/..")) {
+			info_msg("For convenience, you can not add \"..\" directory from the browser view");
+			return;
+		}
+
 		ft = cmus_detect_ft(sel, &ret);
 		if (ft != FILE_TYPE_INVALID) {
 			cmus_add(add, ret, ft, job_type);
@@ -1311,8 +1384,9 @@ static void cmd_win_add_l(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { lib_add_track };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, lib_add_track, 0);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(lib_add_track, JOB_TYPE_LIB);
@@ -1326,8 +1400,9 @@ static void cmd_win_add_p(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { pl_add_track };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, pl_add_track, 0);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(pl_add_track, JOB_TYPE_PL);
@@ -1340,8 +1415,9 @@ static void cmd_win_add_Q(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { play_queue_prepend };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, play_queue_prepend, 1);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 1);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(play_queue_prepend, JOB_TYPE_QUEUE);
@@ -1354,8 +1430,9 @@ static void cmd_win_add_q(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { play_queue_append };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, play_queue_append, 0);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(play_queue_append, JOB_TYPE_QUEUE);
@@ -1598,6 +1675,23 @@ static void cmd_win_pg_up(char *arg)
 	editable_unlock();
 }
 
+static void cmd_win_update_cache(char *arg)
+{
+	struct track_info_selection sel = { .tis = NULL };
+	int flag = parse_flags((const char **)&arg, "f");
+
+	if (cur_view != TREE_VIEW && cur_view != SORTED_VIEW)
+		return;
+
+	editable_lock();
+	view_for_each_sel[cur_view](add_ti, &sel, 0);
+	editable_unlock();
+	if (sel.tis_nr == 0)
+		return;
+	sel.tis[sel.tis_nr] = NULL;
+	cmus_update_tis(sel.tis, sel.tis_nr, flag == 'f');
+}
+
 static void cmd_win_top(char *arg)
 {
 	editable_lock();
@@ -1703,11 +1797,11 @@ found:
 static int count_albums(void)
 {
 	struct artist *artist;
-	struct list_head *item;
+	struct rb_node *tmp1, *tmp2;
 	int count = 0;
 
-	list_for_each_entry(artist, &lib_artist_head, node) {
-		list_for_each(item, &artist->album_head)
+	rb_for_each_entry(artist, tmp1, &lib_artist_root, tree_node) {
+		rb_for_each(tmp2, &artist->album_root)
 			count++;
 	}
 	return count;
@@ -1743,18 +1837,18 @@ static void cmd_lqueue(char *arg)
 		goto unlock;
 
 	r = rand_array(count, nmax);
-	album = to_album(to_artist(lib_artist_head.next)->album_head.next);
+	album = to_album(rb_first(&to_artist(rb_first(&lib_artist_root))->album_root));
 	pos = 0;
 	for (i = 0; i < count; i++) {
 		struct album_list *a;
 
 		while (pos < r[i]) {
 			struct artist *artist = album->artist;
-			if (album->node.next == &artist->album_head) {
-				artist = to_artist(artist->node.next);
-				album = to_album(artist->album_head.next);
+			if (!rb_next(&album->tree_node)) {
+				artist = to_artist(rb_next(&artist->tree_node));
+				album = to_album(rb_first(&artist->album_root));
 			} else {
-				album = to_album(album->node.next);
+				album = to_album(rb_next(&album->tree_node));
 			}
 			pos++;
 		}
@@ -1769,9 +1863,10 @@ static void cmd_lqueue(char *arg)
 		struct list_head *next = item->next;
 		struct album_list *a = container_of(item, struct album_list, node);
 		struct tree_track *t;
-
-		list_for_each_entry(t, &a->album->track_head, node)
-			editable_add(&pq_editable, simple_track_new(tree_track_info(t)));
+		struct rb_node *tmp;
+ 
+		rb_for_each_entry(t, tmp, &a->album->track_root, tree_node)
+			play_queue_append(tree_track_info(t));
 		free(a);
 		item = next;
 	} while (item != &head);
@@ -1779,6 +1874,11 @@ unlock:
 	editable_unlock();
 }
 
+struct track_list {
+	struct list_head node;
+	const struct simple_track *track;
+};
+
 static void cmd_tqueue(char *arg)
 {
 	LIST_HEAD(head);
@@ -1805,13 +1905,14 @@ static void cmd_tqueue(char *arg)
 	item = lib_editable.head.next;
 	pos = 0;
 	for (i = 0; i < count; i++) {
-		struct simple_track *t;
+		struct track_list *t;
 
 		while (pos < r[i]) {
 			item = item->next;
 			pos++;
 		}
-		t = simple_track_new(to_simple_track(item)->info);
+		t = xnew(struct track_list, 1);
+		t->track = to_simple_track(item);
 		list_add_rand(&head, &t->node, i);
 	}
 	free(r);
@@ -1819,8 +1920,9 @@ static void cmd_tqueue(char *arg)
 	item = head.next;
 	do {
 		struct list_head *next = item->next;
-		struct simple_track *t = to_simple_track(item);
-		editable_add(&pq_editable, t);
+		struct track_list *t = container_of(item, struct track_list, node);
+		play_queue_append(t->track->info);
+		free(t);
 		item = next;
 	} while (item != &head);
 unlock:
@@ -1840,6 +1942,11 @@ static int filter_directories(const char *name, const struct stat *s)
 	return S_ISDIR(s->st_mode);
 }
 
+static int filter_executable_files(const char *name, const struct stat *s)
+{
+	return S_ISREG(s->st_mode) && (s->st_mode & 0111);
+}
+
 static int filter_any(const char *name, const struct stat *s)
 {
 	return 1;
@@ -1902,6 +2009,27 @@ static void expand_add(const char *str)
 	}
 }
 
+static void expand_program_paths(const char *str)
+{
+	if (str == NULL)
+		str = "";
+	if (str[0] == '~' || strchr(str, '/'))
+		expand_files(str);
+	else
+		expand_env_path(str, filter_executable_files);
+}
+
+static void expand_program_paths_option(const char *str, const char *opt)
+{
+	expand_program_paths(str);
+
+	if (tabexp.head && opt) {
+		snprintf(expbuf, sizeof(expbuf), "%s=%s", opt, tabexp.head);
+		free(tabexp.head);
+		tabexp.head = xstrdup(expbuf);
+	}
+}
+
 static void expand_load_save(const char *str)
 {
 	int flag = parse_flags(&str, "lp");
@@ -1924,7 +2052,7 @@ static void expand_key_context(const char *str, const char *force)
 	int pos, i, len = strlen(str);
 	char **tails;
 
-	tails = xnew(char *, NR_CTXS + 1);
+	tails = xnew(char *, NR_CTXS);
 	pos = 0;
 	for (i = 0; key_context_names[i]; i++) {
 		int cmp = strncmp(str, key_context_names[i], len);
@@ -1944,10 +2072,10 @@ static void expand_key_context(const char *str, const char *force)
 		free(tails[0]);
 		tails[0] = tmp;
 	}
-	tails[pos] = NULL;
 	snprintf(expbuf, sizeof(expbuf), "%s%s", force, str);
 	tabexp.head = xstrdup(expbuf);
 	tabexp.tails = tails;
+	tabexp.count = pos;
 }
 
 static int get_context(const char *str, int len)
@@ -2040,9 +2168,9 @@ static void expand_bind_args(const char *str)
 
 		snprintf(expbuf, sizeof(expbuf), "%s%s %s", force, key_context_names[c], ks);
 
-		ptr_array_plug(&array);
 		tabexp.head = xstrdup(expbuf);
 		tabexp.tails = array.ptrs;
+		tabexp.count = array.count;
 		return;
 	}
 
@@ -2133,9 +2261,9 @@ static void expand_unbind_args(const char *str)
 
 	snprintf(expbuf, sizeof(expbuf), "%s %s", key_context_names[c], ks);
 
-	ptr_array_plug(&array);
 	tabexp.head = xstrdup(expbuf);
 	tabexp.tails = array.ptrs;
+	tabexp.count = array.count;
 }
 
 static void expand_factivate(const char *str)
@@ -2163,38 +2291,62 @@ static void expand_factivate(const char *str)
 	if (!array.count)
 		return;
 
-	ptr_array_plug(&array);
 	tabexp.head = xstrdup(str);
 	tabexp.tails = array.ptrs;
+	tabexp.count = array.count;
+}
+
+static void expand_fset(const char *str)
+{
+	struct filter_entry *e;
+	PTR_ARRAY(array);
+
+	list_for_each_entry(e, &filters_head, node) {
+		char *line = xnew(char, strlen(e->name) + strlen(e->filter) + 2);
+		sprintf(line, "%s=%s", e->name, e->filter);
+		if (!strncmp(str, line, strlen(str)))
+			ptr_array_add(&array, xstrdup(line + strlen(str)));
+		free(line);
+	}
+	if (!array.count)
+		return;
+
+	tabexp.head = xstrdup(str);
+	tabexp.tails = array.ptrs;
+	tabexp.count = array.count;
 }
 
 static void expand_options(const char *str)
 {
 	struct cmus_opt *opt;
 	int len;
-	char **tails;
+	char **tails, *sep;
 
 	/* tabexp is resetted */
 	len = strlen(str);
-	if (len > 1 && str[len - 1] == '=') {
+	sep = strchr(str, '=');
+	if (len > 1 && sep) {
 		/* expand value */
-		char *var = xstrndup(str, len - 1);
+		char *var = xstrndup(str, sep - str);
 
 		list_for_each_entry(opt, &option_head, node) {
 			if (strcmp(var, opt->name) == 0) {
-				char buf[OPTION_MAX_SIZE];
+				if (str[len - 1] == '=') {
+					char buf[OPTION_MAX_SIZE];
 
-				tails = xnew(char *, 2);
+					tails = xnew(char *, 1);
 
-				buf[0] = 0;
-				opt->get(opt->id, buf);
-				tails[0] = xstrdup(buf);
-				tails[1] = NULL;
+					buf[0] = 0;
+					opt->get(opt->id, buf);
+					tails[0] = xstrdup(buf);
 
-				tabexp.head = xstrdup(str);
-				tabexp.tails = tails;
-				free(var);
-				return;
+					tabexp.head = xstrdup(str);
+					tabexp.tails = tails;
+					tabexp.count = 1;
+				} else if (opt->flags & OPT_PROGRAM_PATH) {
+					expand_program_paths_option(sep + 1, var);
+				}
+				break;
 			}
 		}
 		free(var);
@@ -2202,7 +2354,7 @@ static void expand_options(const char *str)
 		/* expand variable */
 		int pos;
 
-		tails = xnew(char *, nr_options + 1);
+		tails = xnew(char *, nr_options);
 		pos = 0;
 		list_for_each_entry(opt, &option_head, node) {
 			if (strncmp(str, opt->name, len) == 0)
@@ -2217,9 +2369,9 @@ static void expand_options(const char *str)
 				tails[0] = tmp;
 			}
 
-			tails[pos] = NULL;
 			tabexp.head = xstrdup(str);
 			tabexp.tails = tails;
+			tabexp.count = pos;
 		} else {
 			free(tails);
 		}
@@ -2232,7 +2384,7 @@ static void expand_toptions(const char *str)
 	int len, pos;
 	char **tails;
 
-	tails = xnew(char *, nr_options + 1);
+	tails = xnew(char *, nr_options);
 	len = strlen(str);
 	pos = 0;
 	list_for_each_entry(opt, &option_head, node) {
@@ -2242,9 +2394,9 @@ static void expand_toptions(const char *str)
 			tails[pos++] = xstrdup(opt->name + len);
 	}
 	if (pos > 0) {
-		tails[pos] = NULL;
 		tabexp.head = xstrdup(str);
 		tabexp.tails = tails;
+		tabexp.count = pos;
 	} else {
 		free(tails);
 	}
@@ -2288,12 +2440,14 @@ static void expand_colorscheme(const char *str)
 	if (array.count) {
 		ptr_array_sort(&array, strptrcmp);
 
-		ptr_array_plug(&array);
 		tabexp.head = xstrdup(str);
 		tabexp.tails = array.ptrs;
+		tabexp.count = array.count;
 	}
 }
 
+static void expand_commands(const char *str);
+
 /* tab exp }}} */
 
 /* sort by name */
@@ -2307,8 +2461,10 @@ struct command commands[] = {
 	{ "echo",		cmd_echo,	1,-1, NULL,		  0, 0 },
 	{ "factivate",		cmd_factivate,	0, 1, expand_factivate,	  0, 0 },
 	{ "filter",		cmd_filter,	0, 1, NULL,		  0, 0 },
-	{ "fset",		cmd_fset,	1, 1, NULL,		  0, 0 },
+	{ "fset",		cmd_fset,	1, 1, expand_fset,	  0, 0 },
+	{ "help",		cmd_help,	0, 0, NULL,		  0, 0 },
 	{ "invert",		cmd_invert,	0, 0, NULL,		  0, 0 },
+	{ "live-filter",	cmd_live_filter,0, 1, NULL,		  0, CMD_LIVE },
 	{ "load",		cmd_load,	1, 1, expand_load_save,	  0, 0 },
 	{ "lqueue",		cmd_lqueue,	0, 1, NULL,		  0, 0 },
 	{ "mark",		cmd_mark,	0, 1, NULL,		  0, 0 },
@@ -2317,15 +2473,17 @@ struct command commands[] = {
 	{ "player-play",	cmd_p_play,	0, 1, expand_playable,	  0, 0 },
 	{ "player-prev",	cmd_p_prev,	0, 0, NULL,		  0, 0 },
 	{ "player-stop",	cmd_p_stop,	0, 0, NULL,		  0, 0 },
-	{ "prev-view",  	cmd_prev_view,	0, 0, NULL,		  0, 0 },
-	{ "quit",		cmd_quit,	0, 0, NULL,		  0, 0 },
+	{ "prev-view",		cmd_prev_view,	0, 0, NULL,		  0, 0 },
+	{ "push",		cmd_push,	1,-1, expand_commands,	  0, 0 },
+	{ "quit",		cmd_quit,	0, 1, NULL,		  0, 0 },
 	{ "refresh",		cmd_refresh,	0, 0, NULL,		  0, 0 },
-	{ "run",		cmd_run,	1,-1, NULL,		  0, CMD_UNSAFE },
+	{ "run",		cmd_run,	1,-1, expand_program_paths, 0, CMD_UNSAFE },
 	{ "save",		cmd_save,	0, 1, expand_load_save,	  0, CMD_UNSAFE },
 	{ "search-next",	cmd_search_next,0, 0, NULL,		  0, 0 },
 	{ "search-prev",	cmd_search_prev,0, 0, NULL,		  0, 0 },
 	{ "seek",		cmd_seek,	1, 1, NULL,		  0, 0 },
 	{ "set",		cmd_set,	1, 1, expand_options,	  0, 0 },
+	{ "shell",		cmd_shell,	1,-1, expand_program_paths, 0, CMD_UNSAFE },
 	{ "showbind",		cmd_showbind,	1, 1, expand_unbind_args, 0, 0 },
 	{ "shuffle",		cmd_reshuffle,	0, 0, NULL,		  0, 0 },
 	{ "source",		cmd_source,	1, 1, expand_files,	  0, CMD_UNSAFE },
@@ -2333,8 +2491,10 @@ struct command commands[] = {
 	{ "tqueue",		cmd_tqueue,	0, 1, NULL,		  0, 0 },
 	{ "unbind",		cmd_unbind,	1, 1, expand_unbind_args, 0, 0 },
 	{ "unmark",		cmd_unmark,	0, 0, NULL,		  0, 0 },
+	{ "update-cache",	cmd_update_cache,0, 1, NULL,		  0, 0 },
 	{ "view",		cmd_view,	1, 1, NULL,		  0, 0 },
 	{ "vol",		cmd_vol,	1, 2, NULL,		  0, 0 },
+	{ "w",			cmd_save,	0, 1, expand_load_save,	  0, CMD_UNSAFE },
 	{ "win-activate",	cmd_win_activate,0, 0, NULL,		  0, 0 },
 	{ "win-add-l",		cmd_win_add_l,	0, 0, NULL,		  0, 0 },
 	{ "win-add-p",		cmd_win_add_p,	0, 0, NULL,		  0, 0 },
@@ -2353,6 +2513,8 @@ struct command commands[] = {
 	{ "win-top",		cmd_win_top,	0, 0, NULL,		  0, 0 },
 	{ "win-up",		cmd_win_up,	0, 0, NULL,		  0, 0 },
 	{ "win-update",		cmd_win_update,	0, 0, NULL,		  0, 0 },
+	{ "win-update-cache",	cmd_win_update_cache,0, 1, NULL,	  0, 0 },
+	{ "wq",			cmd_quit,	0, 1, NULL,		  0, 0 },
 	{ NULL,			NULL,		0, 0, 0,		  0, 0 }
 };
 
@@ -2363,7 +2525,7 @@ static void expand_commands(const char *str)
 	char **tails;
 
 	/* tabexp is resetted */
-	tails = xnew(char *, sizeof(commands) / sizeof(struct command));
+	tails = xnew(char *, N_ELEMENTS(commands) - 1);
 	len = strlen(str);
 	pos = 0;
 	for (i = 0; commands[i].name; i++) {
@@ -2378,9 +2540,9 @@ static void expand_commands(const char *str)
 			free(tails[0]);
 			tails[0] = tmp;
 		}
-		tails[pos] = NULL;
 		tabexp.head = xstrdup(str);
 		tabexp.tails = tails;
+		tabexp.count = pos;
 	} else {
 		free(tails);
 	}
@@ -2467,7 +2629,7 @@ static void expand_command_line(const char *str)
 	tabexp.head = xstrdup(expbuf);
 }
 
-static void tab_expand(void)
+static void tab_expand(int direction)
 {
 	char *s1, *s2, *tmp;
 	int pos;
@@ -2483,7 +2645,7 @@ static void tab_expand(void)
 	/* tail */
 	s2 = xstrdup(cmdline.line + cmdline.bpos);
 
-	tmp = tabexp_expand(s1, expand_command_line);
+	tmp = tabexp_expand(s1, expand_command_line, direction);
 	if (tmp) {
 		/* tmp.s2 */
 		int l1, l2;
@@ -2498,8 +2660,8 @@ static void tab_expand(void)
 		}
 		sprintf(cmdline.line, "%s%s", tmp, s2);
 		cmdline.bpos = l1;
-		cmdline.cpos = u_strlen(tmp);
-		cmdline.clen = u_strlen(cmdline.line);
+		cmdline.cpos = u_strlen_safe(tmp);
+		cmdline.clen = u_strlen_safe(cmdline.line);
 		free(tmp);
 	}
 	free(s1);
@@ -2512,12 +2674,28 @@ static void reset_tab_expansion(void)
 	arg_expand_cmd = -1;
 }
 
-int run_only_safe_commands;
-
-/* FIXME: parse all arguments */
-void run_command(const char *buf)
+static void cmdline_modified(void)
 {
 	char *cmd, *arg;
+	struct command *c;
+
+	if (!parse_command(cmdline.line, &cmd, &arg))
+		return;
+
+	c = get_command(cmd);
+	if (!c)
+		goto end;
+
+	if (c->flags & CMD_LIVE)
+		run_parsed_command(cmd, arg);
+
+end:
+	free(cmd);
+	free(arg);
+}
+
+int parse_command(const char *buf, char **cmdp, char **argp)
+{
 	int cmd_start, cmd_end, cmd_len;
 	int arg_start, arg_end;
 	int i;
@@ -2527,7 +2705,7 @@ void run_command(const char *buf)
 		i++;
 
 	if (buf[i] == '#')
-		return;
+		return 0;
 
 	cmd_start = i;
 	while (buf[i] && buf[i] != ' ')
@@ -2542,15 +2720,24 @@ void run_command(const char *buf)
 
 	cmd_len = cmd_end - cmd_start;
 	if (cmd_len == 0)
-		return;
+		return 0;
 
-	cmd = xstrndup(buf + cmd_start, cmd_len);
+	*cmdp = xstrndup(buf + cmd_start, cmd_len);
 	if (arg_start == arg_end) {
-		arg = NULL;
+		*argp = NULL;
 	} else {
-		arg = xstrndup(buf + arg_start, arg_end - arg_start);
+		*argp = xstrndup(buf + arg_start, arg_end - arg_start);
 	}
-	i = 0;
+	return 1;
+}
+
+int run_only_safe_commands;
+
+void run_parsed_command(char *cmd, char *arg)
+{
+	int cmd_len = strlen(cmd);
+	int i = 0;
+
 	while (1) {
 		const struct command *c = &commands[i];
 
@@ -2562,7 +2749,7 @@ void run_command(const char *buf)
 			const char *next = commands[i + 1].name;
 			int exact = c->name[cmd_len] == 0;
 
-			if (!exact && next && strncmp(cmd, next, cmd_end - cmd_start) == 0) {
+			if (!exact && next && strncmp(cmd, next, cmd_len) == 0) {
 				error_msg("ambiguous command\n");
 				break;
 			}
@@ -2583,6 +2770,16 @@ void run_command(const char *buf)
 		}
 		i++;
 	}
+}
+
+void run_command(const char *buf)
+{
+	char *cmd, *arg;
+
+	if (!parse_command(buf, &cmd, &arg))
+		return;
+
+	run_parsed_command(cmd, arg);
 	free(arg);
 	free(cmd);
 }
@@ -2614,6 +2811,7 @@ void command_mode_ch(uchar ch)
 		break;
 	case 0x04: // ^D
 		cmdline_delete_ch();
+		cmdline_modified();
 		break;
 	case 0x05: // ^E
 		cmdline_move_end();
@@ -2640,33 +2838,67 @@ void command_mode_ch(uchar ch)
 		break;
 	case 0x0B:
 		cmdline_clear_end();
+		cmdline_modified();
 		break;
 	case 0x09:
-		tab_expand();
+		tab_expand(1);
 		break;
 	case 0x15:
 		cmdline_backspace_to_bol();
+		cmdline_modified();
 		break;
+	case 0x17: // ^W
+		cmdline_backward_delete_word(cmdline_word_delimiters);
+		cmdline_modified();
+		break;
+	case 0x08: // ^H
 	case 127:
 		backspace();
+		cmdline_modified();
 		break;
 	default:
 		cmdline_insert_ch(ch);
+		cmdline_modified();
 	}
 	reset_history_search();
 	if (ch != 0x09)
 		reset_tab_expansion();
 }
 
+void command_mode_escape(int c)
+{
+	switch (c) {
+	case 98:
+		cmdline_backward_word(cmdline_filename_delimiters);
+		break;
+	case 100:
+		cmdline_delete_word(cmdline_filename_delimiters);
+		cmdline_modified();
+		break;
+	case 102:
+		cmdline_forward_word(cmdline_filename_delimiters);
+		break;
+	case 127:
+	case KEY_BACKSPACE:
+		cmdline_backward_delete_word(cmdline_filename_delimiters);
+		cmdline_modified();
+		break;
+	}
+	reset_history_search();
+}
+
 void command_mode_key(int key)
 {
-	reset_tab_expansion();
+	if (key != KEY_BTAB)
+		reset_tab_expansion();
 	switch (key) {
 	case KEY_DC:
 		cmdline_delete_ch();
+		cmdline_modified();
 		break;
 	case KEY_BACKSPACE:
 		backspace();
+		cmdline_modified();
 		break;
 	case KEY_LEFT:
 		cmdline_move_left();
@@ -2703,6 +2935,9 @@ void command_mode_key(int key)
 			}
 		}
 		return;
+	case KEY_BTAB:
+		tab_expand(-1);
+		break;
 	default:
 		d_print("key = %c (%d)\n", key, key);
 	}
@@ -2717,7 +2952,12 @@ void commands_init(void)
 
 void commands_exit(void)
 {
+	view_clear(TREE_VIEW);
+	view_clear(SORTED_VIEW);
+	view_clear(PLAYLIST_VIEW);
+	view_clear(QUEUE_VIEW);
 	history_save(&cmd_history);
+	history_free(&cmd_history);
 	free(cmd_history_filename);
 	tabexp_reset();
 }
