@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2011 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "history.h"
@@ -29,19 +28,33 @@
 
 struct history_entry {
 	struct list_head node;
-	char text[0];
+	char *text;
 };
 
 static struct history_entry *history_entry_new(const char *text)
 {
 	struct history_entry *new;
-	int size = strlen(text) + 1;
-
-	new = xmalloc(sizeof(struct history_entry) + size);
-	memcpy(new->text, text, size);
+	new = xnew(struct history_entry, 1);
+	new->text = xstrdup(text);
 	return new;
 }
 
+static void history_entry_free(struct history_entry *history)
+{
+	free(history->text);
+	free(history);
+}
+
+void history_free(struct history *history)
+{
+	struct list_head *item, *temp;
+	list_for_each_safe(item, temp, &history->head) {
+		struct history_entry *history_entry;
+		history_entry = list_entry(item, struct history_entry, node);
+		history_entry_free(history_entry);
+	}
+}
+
 static int history_add_tail(void *data, const char *line)
 {
 	struct history *history = data;
@@ -77,11 +90,19 @@ void history_save(struct history *history)
 	list_for_each(item, &history->head) {
 		struct history_entry *history_entry;
 		const char nl = '\n';
+		ssize_t rc;
 
 		history_entry = list_entry(item, struct history_entry, node);
-		write(fd, history_entry->text, strlen(history_entry->text));
-		write(fd, &nl, 1);
+
+		rc = write(fd, history_entry->text, strlen(history_entry->text));
+		if (rc == -1)
+			goto out;
+
+		rc = write(fd, &nl, 1);
+		if (rc == -1)
+			goto out;
 	}
+out:
 	close(fd);
 }
 
@@ -103,7 +124,7 @@ void history_add_line(struct history *history, const char *line)
 		hentry = container_of(item, struct history_entry, node);
 		if (strcmp(hentry->text, new->text) == 0) {
 			list_del(item);
-			free(hentry);
+			history_entry_free(hentry);
 			history->lines--;
 		}
 		item = next;
@@ -117,7 +138,7 @@ void history_add_line(struct history *history, const char *line)
 		node = history->head.prev;
 		list_del(node);
 		hentry = list_entry(node, struct history_entry, node);
-		free(hentry);
+		history_entry_free(hentry);
 		history->lines--;
 	}
 }
