@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2011 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "input.h"
@@ -23,23 +22,32 @@
 #include "http.h"
 #include "xmalloc.h"
 #include "file.h"
+#include "path.h"
 #include "utils.h"
 #include "cmus.h"
+#include "options.h"
 #include "list.h"
+#include "mergesort.h"
 #include "misc.h"
 #include "debug.h"
 #include "ui_curses.h"
+#ifdef HAVE_CONFIG
 #include "config/libdir.h"
+#endif
 
 #include <unistd.h>
 #include <string.h>
 #include <errno.h>
+#include <stdio.h>
 #include <stdarg.h>
+#include <sys/select.h>
+#include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <dirent.h>
 #include <dlfcn.h>
+#include <strings.h>
 
 struct input_plugin {
 	const struct input_plugin_ops *ops;
@@ -51,6 +59,12 @@ struct input_plugin {
 
 	/* cached duration, -1 = unset */
 	int duration;
+	/* cached bitrate, -1 = unset */
+	long bitrate;
+	/* cached codec, NULL = unset */
+	char *codec;
+	/* cached codec_profile, NULL = unset */
+	char *codec_profile;
 
 	/*
 	 * pcm is converted to 16-bit signed little-endian stereo
@@ -71,9 +85,11 @@ struct ip {
 	char *name;
 	void *handle;
 
+	int priority;
 	const char * const *extensions;
 	const char * const *mime_types;
 	const struct input_plugin_ops *ops;
+	const char * const *options;
 };
 
 static const char * const plugin_dir = LIBDIR "/cmus/ip";
@@ -89,36 +105,20 @@ static const char *pl_mime_types[] = {
 	"audio/x-mpegurl"
 };
 
-static const char *get_extension(const char *filename)
+static const struct input_plugin_ops *get_ops_by_extension(const char *ext, struct list_head **headp)
 {
-	const char *ext;
+	struct list_head *node = *headp;
 
-	ext = filename + strlen(filename) - 1;
-	while (ext >= filename && *ext != '/') {
-		if (*ext == '.') {
-			ext++;
-			return ext;
-		}
-		ext--;
-	}
-	return NULL;
-}
-
-static const struct input_plugin_ops *get_ops_by_filename(const char *filename)
-{
-	struct ip *ip;
-	const char *ext;
-
-	ext = get_extension(filename);
-	if (ext == NULL)
-		return NULL;
-	list_for_each_entry(ip, &ip_head, node) {
+	for (node = node->next; node != &ip_head; node = node->next) {
+		struct ip *ip = list_entry(node, struct ip, node);
 		const char * const *exts = ip->extensions;
 		int i;
 
 		for (i = 0; exts[i]; i++) {
-			if (strcasecmp(ext, exts[i]) == 0)
+			if (strcasecmp(ext, exts[i]) == 0 || strcmp("*", exts[i]) == 0) {
+				*headp = node;
 				return ip->ops;
+			}
 		}
 	}
 	return NULL;
@@ -140,88 +140,103 @@ static const struct input_plugin_ops *get_ops_by_mime_type(const char *mime_type
 	return NULL;
 }
 
-static int do_http_get(const char *uri, struct http_header **headersp, int *codep, char **reasonp)
+static void keyvals_add_basic_auth(struct growing_keyvals *c,
+				   const char *user,
+				   const char *pass,
+				   const char *header)
 {
-	struct http_uri u;
-	struct http_header *h;
-	int sock, i, rc, code;
-	char *reason;
+	char buf[256];
+	char *encoded;
 
-	*headersp = NULL;
-	*codep = -1;
-	*reasonp = NULL;
-
-	if (http_parse_uri(uri, &u))
-		return -IP_ERROR_INVALID_URI;
-
-/* 	d_print("%s -> '%s':'%s'@'%s':%d'%s'\n", uri, user, pass, host, port, path); */
-
-	sock = http_open(u.host, u.port, http_connection_timeout);
-	if (sock == -1) {
-		http_free_uri(&u);
-		return -IP_ERROR_ERRNO;
+	snprintf(buf, sizeof(buf), "%s:%s", user, pass);
+	encoded = base64_encode(buf);
+	if (encoded == NULL) {
+		d_print("couldn't base64 encode '%s'\n", buf);
+	} else {
+		snprintf(buf, sizeof(buf), "Basic %s", encoded);
+		free(encoded);
+		keyvals_add(c, header, xstrdup(buf));
 	}
-
-	h = xnew(struct http_header, 5);
-	i = 0;
-	h[i].key = xstrdup("Host");
-	h[i].val = xstrdup(u.host);
-	i++;
-	h[i].key = xstrdup("User-Agent");
-	h[i].val = xstrdup("cmus/" VERSION);
-	i++;
-	h[i].key = xstrdup("Icy-MetaData");
-	h[i].val = xstrdup("1");
-	i++;
-	if (u.user && u.pass) {
-		char buf[256];
-		char *encoded;
-
-		snprintf(buf, sizeof(buf), "%s:%s", u.user, u.pass);
-		encoded = base64_encode(buf);
-		if (encoded == NULL) {
-			d_print("couldn't base64 encode '%s'\n", buf);
-		} else {
-			snprintf(buf, sizeof(buf), "Basic %s", encoded);
-			free(encoded);
-			h[i].key = xstrdup("Authorization");
-			h[i].val = xstrdup(buf);
-			i++;
-		}
-	}
-	h[i].key = NULL;
-	h[i].val = NULL;
-	i++;
-
-	rc = http_get(sock, u.path, h, &code, &reason, headersp, http_read_timeout);
-	http_headers_free(h);
-	http_free_uri(&u);
-	switch (rc) {
-	case -1:
-		d_print("error: %s\n", strerror(errno));
-		close(sock);
-		return -IP_ERROR_ERRNO;
-	case -2:
-		d_print("error parsing HTTP response\n");
-		close(sock);
-		return -IP_ERROR_HTTP_RESPONSE;
-	}
-	d_print("HTTP response: %d %s\n", code, reason);
-	if (code != 200) {
-		*codep = code;
-		*reasonp = reason;
-		close(sock);
-		return -IP_ERROR_HTTP_STATUS;
-	}
-	free(reason);
-	return sock;
 }
 
-static int setup_remote(struct input_plugin *ip, const struct http_header *headers, int sock)
+static int do_http_get(struct http_get *hg, const char *uri, int redirections)
+{
+	GROWING_KEYVALS(h);
+	int i, rc;
+	const char *val;
+	char *redirloc;
+
+	d_print("%s\n", uri);
+
+	hg->headers = NULL;
+	hg->reason = NULL;
+	hg->code = -1;
+	hg->fd = -1;
+	if (http_parse_uri(uri, &hg->uri))
+		return -IP_ERROR_INVALID_URI;
+
+	if (http_open(hg, http_connection_timeout))
+		return -IP_ERROR_ERRNO;
+
+	keyvals_add(&h, "Host", xstrdup(hg->uri.host));
+	if (hg->proxy && hg->proxy->user && hg->proxy->pass)
+		keyvals_add_basic_auth(&h, hg->proxy->user, hg->proxy->pass, "Proxy-Authorization");
+	keyvals_add(&h, "User-Agent", xstrdup("cmus/" VERSION));
+	keyvals_add(&h, "Icy-MetaData", xstrdup("1"));
+	if (hg->uri.user && hg->uri.pass)
+		keyvals_add_basic_auth(&h, hg->uri.user, hg->uri.pass, "Authorization");
+	keyvals_terminate(&h);
+
+	rc = http_get(hg, h.keyvals, http_read_timeout);
+	keyvals_free(h.keyvals);
+	switch (rc) {
+	case -1:
+		return -IP_ERROR_ERRNO;
+	case -2:
+		return -IP_ERROR_HTTP_RESPONSE;
+	}
+
+	d_print("HTTP response: %d %s\n", hg->code, hg->reason);
+	for (i = 0; hg->headers[i].key != NULL; i++)
+		d_print("  %s: %s\n", hg->headers[i].key, hg->headers[i].val);
+
+	switch (hg->code) {
+	case 200: /* OK */
+		return 0;
+	/*
+	 * 3xx Codes (Redirections)
+	 *     unhandled: 300 Multiple Choices
+	 */
+	case 301: /* Moved Permanently */
+	case 302: /* Found */
+	case 303: /* See Other */
+	case 307: /* Temporary Redirect */
+		val = keyvals_get_val(hg->headers, "location");
+		if (!val)
+			return -IP_ERROR_HTTP_RESPONSE;
+
+		redirections++;
+		if (redirections > 2)
+			return -IP_ERROR_HTTP_REDIRECT_LIMIT;
+
+		redirloc = xstrdup(val);
+		http_get_free(hg);
+		close(hg->fd);
+
+		rc = do_http_get(hg, redirloc, redirections);
+
+		free(redirloc);
+		return rc;
+	default:
+		return -IP_ERROR_HTTP_STATUS;
+	}
+}
+
+static int setup_remote(struct input_plugin *ip, const struct keyval *headers, int sock)
 {
 	const char *val;
 
-	val = http_headers_get_value(headers, "Content-Type");
+	val = keyvals_get_val(headers, "Content-Type");
 	if (val) {
 		d_print("Content-Type: %s\n", val);
 		ip->ops = get_ops_by_mime_type(val);
@@ -243,9 +258,9 @@ static int setup_remote(struct input_plugin *ip, const struct http_header *heade
 	}
 
 	ip->data.fd = sock;
-	ip->data.metadata = (char *)xmalloc(16 * 255 + 1);
+	ip->data.metadata = xnew(char, 16 * 255 + 1);
 
-	val = http_headers_get_value(headers, "icy-metaint");
+	val = keyvals_get_val(headers, "icy-metaint");
 	if (val) {
 		long int lint;
 
@@ -254,99 +269,178 @@ static int setup_remote(struct input_plugin *ip, const struct http_header *heade
 			d_print("metaint: %d\n", ip->data.metaint);
 		}
 	}
+
+	val = keyvals_get_val(headers, "icy-name");
+	if (val)
+		ip->data.icy_name = xstrdup(val);
+
+	val = keyvals_get_val(headers, "icy-genre");
+	if (val)
+		ip->data.icy_genre = xstrdup(val);
+
+	val = keyvals_get_val(headers, "icy-url");
+	if (val)
+		ip->data.icy_url = xstrdup(val);
+
 	return 0;
 }
 
-static int handle_line(void *data, const char *line)
-{
-	char **firstp = data;
+struct read_playlist_data {
+	struct input_plugin *ip;
+	int rc;
+	int count;
+};
 
-	*firstp = xstrdup(line);
-	/* ignore other lines */
+static int handle_line(void *data, const char *uri)
+{
+	struct read_playlist_data *rpd = data;
+	struct http_get hg;
+
+	rpd->count++;
+	rpd->rc = do_http_get(&hg, uri, 0);
+	if (rpd->rc) {
+		rpd->ip->http_code = hg.code;
+		rpd->ip->http_reason = hg.reason;
+		if (hg.fd >= 0)
+			close(hg.fd);
+
+		hg.reason = NULL;
+		http_get_free(&hg);
+		return 0;
+	}
+
+	rpd->rc = setup_remote(rpd->ip, hg.headers, hg.fd);
+	http_get_free(&hg);
 	return 1;
 }
 
 static int read_playlist(struct input_plugin *ip, int sock)
 {
-	struct http_header *headers;
-	char *body, *reason, *first;
-	int rc, code;
+	struct read_playlist_data rpd = { ip, 0, 0 };
+	char *body;
+	size_t size;
 
-	rc = http_read_body(sock, &body, http_read_timeout);
+	body = http_read_body(sock, &size, http_read_timeout);
 	close(sock);
-	if (rc)
+	if (!body)
 		return -IP_ERROR_ERRNO;
 
-	/* get only first URL from the playlist */
-	first = NULL;
-	cmus_playlist_for_each(body, strlen(body), 0, handle_line, &first);
+	cmus_playlist_for_each(body, size, 0, handle_line, &rpd);
 	free(body);
-
-	if (first == NULL) {
+	if (!rpd.count) {
 		d_print("empty playlist\n");
-		return -IP_ERROR_HTTP_RESPONSE;
+		rpd.rc = -IP_ERROR_HTTP_RESPONSE;
 	}
-
-	sock = do_http_get(first, &headers, &code, &reason);
-	free(first);
-	if (sock < 0) {
-		ip->http_code = code;
-		ip->http_reason = reason;
-		/* URI in the _playlist_ is invalid, not our fault */
-		if (sock == -IP_ERROR_INVALID_URI)
-			sock = -IP_ERROR_HTTP_RESPONSE;
-		return sock;
-	}
-
-	rc = setup_remote(ip, headers, sock);
-	http_headers_free(headers);
-	return rc;
+	return rpd.rc;
 }
 
 static int open_remote(struct input_plugin *ip)
 {
 	struct input_plugin_data *d = &ip->data;
-	char *reason;
-	int sock, rc, code;
-	struct http_header *headers;
+	struct http_get hg;
 	const char *val;
+	int rc;
 
-	sock = do_http_get(d->filename, &headers, &code, &reason);
-	if (sock < 0) {
-		ip->http_code = code;
-		ip->http_reason = reason;
-		return sock;
+	rc = do_http_get(&hg, d->filename, 0);
+	if (rc) {
+		ip->http_code = hg.code;
+		ip->http_reason = hg.reason;
+		hg.reason = NULL;
+		http_get_free(&hg);
+		return rc;
 	}
 
-	val = http_headers_get_value(headers, "Content-Type");
+	val = keyvals_get_val(hg.headers, "Content-Type");
 	if (val) {
 		int i;
 
-		for (i = 0; i < sizeof(pl_mime_types) / sizeof(pl_mime_types[0]); i++) {
+		for (i = 0; i < N_ELEMENTS(pl_mime_types); i++) {
 			if (!strcasecmp(val, pl_mime_types[i])) {
 				d_print("Content-Type: %s\n", val);
-				http_headers_free(headers);
-				return read_playlist(ip, sock);
+				http_get_free(&hg);
+				return read_playlist(ip, hg.fd);
 			}
 		}
 	}
 
-	rc = setup_remote(ip, headers, sock);
-	http_headers_free(headers);
+	rc = setup_remote(ip, hg.headers, hg.fd);
+	http_get_free(&hg);
 	return rc;
 }
 
+static void ip_init(struct input_plugin *ip, char *filename)
+{
+	const struct input_plugin t = {
+		.http_code          = -1,
+		.pcm_convert_scale  = -1,
+		.duration           = -1,
+		.bitrate            = -1,
+		.data = {
+			.fd         = -1,
+			.filename   = filename,
+			.remote     = is_http_url(filename),
+			.channel_map = CHANNEL_MAP_INIT
+		}
+	};
+	*ip = t;
+}
+
+static void ip_reset(struct input_plugin *ip, int close_fd)
+{
+	int fd = ip->data.fd;
+	free(ip->data.metadata);
+	ip_init(ip, ip->data.filename);
+	if (fd != -1) {
+		if (close_fd)
+			close(fd);
+		else {
+			lseek(fd, 0, SEEK_SET);
+			ip->data.fd = fd;
+		}
+	}
+}
+
 static int open_file(struct input_plugin *ip)
 {
-	ip->ops = get_ops_by_filename(ip->data.filename);
-	if (ip->ops == NULL)
+	const struct input_plugin_ops *ops;
+	struct list_head *head = &ip_head;
+	const char *ext;
+	int rc = 0;
+
+	ext = get_extension(ip->data.filename);
+	if (!ext)
 		return -IP_ERROR_UNRECOGNIZED_FILE_TYPE;
+
+	ops = get_ops_by_extension(ext, &head);
+	if (!ops)
+		return -IP_ERROR_UNRECOGNIZED_FILE_TYPE;
+
 	ip->data.fd = open(ip->data.filename, O_RDONLY);
-	if (ip->data.fd == -1) {
-		ip->ops = NULL;
+	if (ip->data.fd == -1)
 		return -IP_ERROR_ERRNO;
+
+	while (1) {
+		ip->ops = ops;
+		rc = ip->ops->open(&ip->data);
+		if (rc != -IP_ERROR_UNSUPPORTED_FILE_TYPE)
+			break;
+
+		ops = get_ops_by_extension(ext, &head);
+		if (!ops)
+			break;
+
+		ip_reset(ip, 0);
+		d_print("fallback: try next plugin for `%s'\n", ip->data.filename);
 	}
-	return 0;
+
+	return rc;
+}
+
+static int sort_ip(const struct list_head *a_, const struct list_head *b_)
+{
+	const struct ip *a = list_entry(a_, struct ip, node);
+	const struct ip *b = list_entry(b_, struct ip, node);
+	return b->priority - a->priority;
 }
 
 void ip_load_plugins(void)
@@ -359,12 +453,12 @@ void ip_load_plugins(void)
 		error_msg("couldn't open directory `%s': %s", plugin_dir, strerror(errno));
 		return;
 	}
-	while ((d = readdir(dir)) != NULL) {
+	while ((d = (struct dirent *) readdir(dir)) != NULL) {
 		char filename[256];
 		struct ip *ip;
 		void *so;
 		char *ext;
-		const char *sym;
+		const int *priority_ptr;
 
 		if (d->d_name[0] == '.')
 			continue;
@@ -378,48 +472,34 @@ void ip_load_plugins(void)
 
 		so = dlopen(filename, RTLD_NOW);
 		if (so == NULL) {
-			error_msg("%s", dlerror());
+			d_print("%s: %s\n", filename, dlerror());
 			continue;
 		}
 
 		ip = xnew(struct ip, 1);
 
-		sym = "ip_extensions";
-		if (!(ip->extensions = dlsym(so, sym)))
-			goto sym_err;
-
-		sym = "ip_mime_types";
-		if (!(ip->mime_types = dlsym(so, sym)))
-			goto sym_err;
-
-		sym = "ip_ops";
-		if (!(ip->ops = dlsym(so, sym)))
-			goto sym_err;
+		priority_ptr = dlsym(so, "ip_priority");
+		ip->extensions = dlsym(so, "ip_extensions");
+		ip->mime_types = dlsym(so, "ip_mime_types");
+		ip->ops = dlsym(so, "ip_ops");
+		ip->options = dlsym(so, "ip_options");
+		if (!priority_ptr || !ip->extensions || !ip->mime_types || !ip->ops || !ip->options) {
+			error_msg("%s: missing symbol", filename);
+			free(ip);
+			dlclose(so);
+			continue;
+		}
+		ip->priority = *priority_ptr;
 
 		ip->name = xstrndup(d->d_name, ext - d->d_name);
 		ip->handle = so;
 
 		list_add_tail(&ip->node, &ip_head);
-		continue;
-sym_err:
-		error_msg("%s: symbol %s not found", filename, sym);
-		free(ip);
-		dlclose(so);
 	}
+	list_mergesort(&ip_head, sort_ip);
 	closedir(dir);
 }
 
-static void ip_init(struct input_plugin *ip, char *filename)
-{
-	memset(ip, 0, sizeof(*ip));
-	ip->http_code = -1;
-	ip->pcm_convert_scale = -1;
-	ip->duration = -1;
-	ip->data.fd = -1;
-	ip->data.filename = filename;
-	ip->data.remote = is_url(filename);
-}
-
 struct input_plugin *ip_new(const char *filename)
 {
 	struct input_plugin *ip = xnew(struct input_plugin, 1);
@@ -442,27 +522,26 @@ int ip_open(struct input_plugin *ip)
 
 	BUG_ON(ip->open);
 
-	/* set fd and ops */
+	/* set fd and ops, call ops->open */
 	if (ip->data.remote) {
 		rc = open_remote(ip);
+		if (rc == 0)
+			rc = ip->ops->open(&ip->data);
 	} else {
-		rc = open_file(ip);
+		if (is_cdda_url(ip->data.filename)) {
+			ip->ops = get_ops_by_mime_type("x-content/audio-cdda");
+			rc = ip->ops ? ip->ops->open(&ip->data) : 1;
+		} else if (is_cue_url(ip->data.filename)) {
+			ip->ops = get_ops_by_mime_type("application/x-cue");
+			rc = ip->ops ? ip->ops->open(&ip->data) : 1;
+		} else
+			rc = open_file(ip);
 	}
 
 	if (rc) {
 		d_print("opening `%s' failed: %d %s\n", ip->data.filename, rc,
 				rc == -1 ? strerror(errno) : "");
-		return rc;
-	}
-
-	rc = ip->ops->open(&ip->data);
-	if (rc) {
-		d_print("opening file `%s' failed: %d %s\n", ip->data.filename, rc,
-				rc == -1 ? strerror(errno) : "");
-		if (ip->data.fd != -1)
-			close(ip->data.fd);
-		free(ip->data.metadata);
-		ip_init(ip, ip->data.filename);
+		ip_reset(ip, 1);
 		return rc;
 	}
 	ip->open = 1;
@@ -506,6 +585,9 @@ int ip_close(struct input_plugin *ip)
 	if (ip->data.fd != -1)
 		close(ip->data.fd);
 	free(ip->data.metadata);
+	free(ip->data.icy_name);
+	free(ip->data.icy_genre);
+	free(ip->data.icy_url);
 	free(ip->http_reason);
 
 	ip_init(ip, ip->data.filename);
@@ -559,7 +641,7 @@ int ip_read(struct input_plugin *ip, char *buffer, int count)
 		return rc;
 	}
 
-	BUG_ON((rc & ~((unsigned int)sf_get_frame_size(ip->data.sf) - 1U)) != rc);
+	BUG_ON(rc % sf_get_frame_size(ip->data.sf) != 0);
 
 	sample_size = sf_get_sample_size(ip->data.sf);
 	if (ip->pcm_convert_in_place != NULL)
@@ -583,7 +665,36 @@ int ip_seek(struct input_plugin *ip, double offset)
 
 int ip_read_comments(struct input_plugin *ip, struct keyval **comments)
 {
-	return ip->ops->read_comments(&ip->data, comments);
+	struct keyval *kv = NULL;
+	int rc;
+
+	rc = ip->ops->read_comments(&ip->data, &kv);
+
+	if (ip->data.remote) {
+		GROWING_KEYVALS(c);
+
+		if (kv) {
+			keyvals_init(&c, kv);
+			keyvals_free(kv);
+		}
+
+		if (ip->data.icy_name && !keyvals_get_val_growing(&c, "title"))
+			keyvals_add(&c, "title", xstrdup(ip->data.icy_name));
+
+		if (ip->data.icy_genre && !keyvals_get_val_growing(&c, "genre"))
+			keyvals_add(&c, "genre", xstrdup(ip->data.icy_genre));
+
+		if (ip->data.icy_url && !keyvals_get_val_growing(&c, "comment"))
+			keyvals_add(&c, "comment", xstrdup(ip->data.icy_url));
+
+		keyvals_terminate(&c);
+
+		kv = c.keyvals;
+	}
+
+	*comments = kv;
+
+	return ip->data.remote ? 0 : rc;
 }
 
 int ip_duration(struct input_plugin *ip)
@@ -597,12 +708,52 @@ int ip_duration(struct input_plugin *ip)
 	return ip->duration;
 }
 
+int ip_bitrate(struct input_plugin *ip)
+{
+	if (ip->data.remote)
+		return -1;
+	if (ip->bitrate == -1)
+		ip->bitrate = ip->ops->bitrate(&ip->data);
+	if (ip->bitrate < 0)
+		return -1;
+	return ip->bitrate;
+}
+
+int ip_current_bitrate(struct input_plugin *ip)
+{
+	return ip->ops->bitrate_current(&ip->data);
+}
+
+char *ip_codec(struct input_plugin *ip)
+{
+	if (ip->data.remote)
+		return NULL;
+	if (!ip->codec)
+		ip->codec = ip->ops->codec(&ip->data);
+	return ip->codec;
+}
+
+char *ip_codec_profile(struct input_plugin *ip)
+{
+	if (ip->data.remote)
+		return NULL;
+	if (!ip->codec_profile)
+		ip->codec_profile = ip->ops->codec_profile(&ip->data);
+	return ip->codec_profile;
+}
+
 sample_format_t ip_get_sf(struct input_plugin *ip)
 {
 	BUG_ON(!ip->open);
 	return ip->data.sf;
 }
 
+void ip_get_channel_map(struct input_plugin *ip, channel_position_t *channel_map)
+{
+	BUG_ON(!ip->open);
+	channel_map_copy(channel_map, ip->data.channel_map);
+}
+
 const char *ip_get_filename(struct input_plugin *ip)
 {
 	return ip->data.filename;
@@ -634,6 +785,64 @@ int ip_eof(struct input_plugin *ip)
 	return ip->eof;
 }
 
+static struct ip *find_plugin(int idx)
+{
+	struct ip *ip;
+
+	list_for_each_entry(ip, &ip_head, node) {
+		if (idx == 0)
+			return ip;
+		idx--;
+	}
+	return NULL;
+}
+
+static void option_error(int rc)
+{
+	char *msg = ip_get_error_msg(NULL, rc, "setting option");
+	error_msg("%s", msg);
+	free(msg);
+}
+
+static void set_ip_option(unsigned int id, const char *val)
+{
+	const struct ip *ip = find_plugin(id >> 16);
+	int rc;
+
+	rc = ip->ops->set_option(id & 0xffff, val);
+	if (rc)
+		option_error(rc);
+}
+
+static void get_ip_option(unsigned int id, char *buf)
+{
+	const struct ip *ip = find_plugin(id >> 16);
+	char *val = NULL;
+
+	ip->ops->get_option(id & 0xffff, &val);
+	if (val) {
+		strcpy(buf, val);
+		free(val);
+	}
+}
+
+void ip_add_options(void)
+{
+	struct ip *ip;
+	unsigned int iid, pid = 0;
+	char key[64];
+
+	list_for_each_entry(ip, &ip_head, node) {
+		for (iid = 0; ip->options[iid]; iid++) {
+			snprintf(key, sizeof(key), "input.%s.%s",
+					ip->name,
+					ip->options[iid]);
+			option_add(xstrdup(key), (pid << 16) | iid, get_ip_option, set_ip_option, NULL, 0);
+		}
+		pid++;
+	}
+}
+
 char *ip_get_error_msg(struct input_plugin *ip, int rc, const char *arg)
 {
 	char buffer[1024];
@@ -646,6 +855,10 @@ char *ip_get_error_msg(struct input_plugin *ip, int rc, const char *arg)
 		snprintf(buffer, sizeof(buffer),
 				"%s: unrecognized filename extension", arg);
 		break;
+	case IP_ERROR_UNSUPPORTED_FILE_TYPE:
+		snprintf(buffer, sizeof(buffer),
+				"%s: unsupported file format", arg);
+		break;
 	case IP_ERROR_FUNCTION_NOT_SUPPORTED:
 		snprintf(buffer, sizeof(buffer),
 				"%s: function not supported", arg);
@@ -663,6 +876,12 @@ char *ip_get_error_msg(struct input_plugin *ip, int rc, const char *arg)
 				"%s: input plugin doesn't support the sample format",
 				arg);
 		break;
+	case IP_ERROR_WRONG_DISC:
+		snprintf(buffer, sizeof(buffer), "%s: wrong disc inserted, aborting!", arg);
+		break;
+	case IP_ERROR_NO_DISC:
+		snprintf(buffer, sizeof(buffer), "%s: could not read disc", arg);
+		break;
 	case IP_ERROR_HTTP_RESPONSE:
 		snprintf(buffer, sizeof(buffer), "%s: invalid HTTP response", arg);
 		break;
@@ -672,6 +891,13 @@ char *ip_get_error_msg(struct input_plugin *ip, int rc, const char *arg)
 		ip->http_reason = NULL;
 		ip->http_code = -1;
 		break;
+	case IP_ERROR_HTTP_REDIRECT_LIMIT:
+		snprintf(buffer, sizeof(buffer), "%s: too many HTTP redirections", arg);
+		break;
+	case IP_ERROR_NOT_OPTION:
+		snprintf(buffer, sizeof(buffer),
+				"%s: no such option", arg);
+		break;
 	case IP_ERROR_INTERNAL:
 		snprintf(buffer, sizeof(buffer), "%s: internal error", arg);
 		break;
@@ -717,7 +943,7 @@ void ip_dump_plugins(void)
 
 	printf("Input Plugins: %s\n", plugin_dir);
 	list_for_each_entry(ip, &ip_head, node) {
-		printf("  %s:\n    File Types:", ip->name);
+		printf("  %s:\n    Priority: %d\n    File Types:", ip->name, ip->priority);
 		for (i = 0; ip->extensions[i]; i++)
 			printf(" %s", ip->extensions[i]);
 		printf("\n    MIME Types:");
