@@ -1,5 +1,19 @@
 /*
+ * Copyright 2008-2011 Various Authors
  * Copyright 2006 Timo Hirvonen
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "options.h"
@@ -9,7 +23,6 @@
 #include "player.h"
 #include "buffer.h"
 #include "ui_curses.h"
-#include "format_print.h"
 #include "cmus.h"
 #include "misc.h"
 #include "lib.h"
@@ -21,10 +34,18 @@
 #include "file.h"
 #include "prog.h"
 #include "output.h"
+#include "input.h"
+#ifdef HAVE_CONFIG
 #include "config/datadir.h"
+#endif
+#include "track_info.h"
+#include "cache.h"
+#include "debug.h"
+#include "discid.h"
 
 #include <stdio.h>
 #include <errno.h>
+#include <strings.h>
 
 #if defined(__sun__)
 #include <ncurses.h>
@@ -34,17 +55,23 @@
 
 /* initialized option variables */
 
+char *cdda_device = NULL;
 char *output_plugin = NULL;
 char *status_display_program = NULL;
 char *server_password;
-int auto_reshuffle = 0;
+int auto_reshuffle = 1;
 int confirm_run = 1;
+int resume_cmus = 0;
 int show_hidden = 0;
+int show_current_bitrate = 0;
 int show_remaining_time = 0;
 int set_term_title = 1;
+int wrap_search = 1;
 int play_library = 1;
 int repeat = 0;
 int shuffle = 0;
+int display_artist_sort_name;
+int smart_artist_sort = 1;
 
 int colors[NR_COLORS] = {
 	-1,
@@ -79,8 +106,10 @@ int colors[NR_COLORS] = {
 
 /* uninitialized option variables */
 char *track_win_format = NULL;
+char *track_win_format_va = NULL;
 char *track_win_alt_format = NULL;
 char *list_win_format = NULL;
+char *list_win_format_va = NULL;
 char *list_win_alt_format = NULL;
 char *current_format = NULL;
 char *current_alt_format = NULL;
@@ -145,13 +174,27 @@ enum format_id {
 	FMT_TRACKWIN_ALT,
 	FMT_CURRENT,
 	FMT_PLAYLIST,
+	FMT_PLAYLIST_VA,
 	FMT_TITLE,
-	FMT_TRACKWIN
+	FMT_TRACKWIN,
+	FMT_TRACKWIN_VA,
+
+	NR_FMTS
 };
-#define NR_FMTS 8
 
 /* callbacks for normal options {{{ */
 
+static void get_device(unsigned int id, char *buf)
+{
+	strcpy(buf, cdda_device);
+}
+
+static void set_device(unsigned int id, const char *buf)
+{
+	free(cdda_device);
+	cdda_device = expand_filename(buf);
+}
+
 #define SECOND_SIZE (44100 * 16 / 8 * 2)
 static void get_buffer_seconds(unsigned int id, char *buf)
 {
@@ -162,7 +205,7 @@ static void set_buffer_seconds(unsigned int id, const char *buf)
 {
 	int sec;
 
-	if (parse_int(buf, 1, 20, &sec))
+	if (parse_int(buf, 1, 300, &sec))
 		player_set_buffer_chunks((sec * SECOND_SIZE + CHUNK_SIZE / 2) / CHUNK_SIZE);
 }
 
@@ -177,28 +220,42 @@ static void set_id3_default_charset(unsigned int id, const char *buf)
 	id3_default_charset = xstrdup(buf);
 }
 
-static const char * const valid_sort_keys[] = {
-	"artist",
-	"album",
-	"title",
-	"tracknumber",
-	"discnumber",
-	"date",
-	"genre",
-	"filename",
-	"albumartist",
-	NULL
+static const struct {
+	const char *str;
+	sort_key_t key;
+} sort_key_map[] = {
+	{ "artist",		SORT_ARTIST		},
+	{ "album",		SORT_ALBUM		},
+	{ "title",		SORT_TITLE		},
+	{ "tracknumber",	SORT_TRACKNUMBER	},
+	{ "discnumber",		SORT_DISCNUMBER		},
+	{ "date",		SORT_DATE		},
+	{ "originaldate",	SORT_ORIGINALDATE	},
+	{ "genre",		SORT_GENRE		},
+	{ "comment",		SORT_COMMENT		},
+	{ "albumartist",	SORT_ALBUMARTIST	},
+	{ "filename",		SORT_FILENAME		},
+	{ "filemtime",		SORT_FILEMTIME		},
+	{ "rg_track_gain",	SORT_RG_TRACK_GAIN	},
+	{ "rg_track_peak",	SORT_RG_TRACK_PEAK	},
+	{ "rg_album_gain",	SORT_RG_ALBUM_GAIN	},
+	{ "rg_album_peak",	SORT_RG_ALBUM_PEAK	},
+	{ "bitrate",		SORT_BITRATE		},
+	{ "codec",		SORT_CODEC		},
+	{ "codec_profile",	SORT_CODEC_PROFILE	},
+	{ "media",		SORT_MEDIA		},
+	{ NULL,                 SORT_INVALID            }
 };
 
-static const char **parse_sort_keys(const char *value)
+static sort_key_t *parse_sort_keys(const char *value)
 {
-	const char **keys;
+	sort_key_t *keys;
 	const char *s, *e;
 	int size = 4;
 	int pos = 0;
 
 	size = 4;
-	keys = xnew(const char *, size);
+	keys = xnew(sort_key_t, size);
 
 	s = value;
 	while (1) {
@@ -223,26 +280,55 @@ static const char **parse_sort_keys(const char *value)
 		s = e;
 
 		for (i = 0; ; i++) {
-			if (valid_sort_keys[i] == NULL) {
+			if (sort_key_map[i].str == NULL) {
 				error_msg("invalid sort key '%s'", buf);
 				free(keys);
 				return NULL;
 			}
 
-			if (strcmp(buf, valid_sort_keys[i]) == 0)
+			if (strcmp(buf, sort_key_map[i].str) == 0)
 				break;
 		}
-
 		if (pos == size - 1) {
 			size *= 2;
-			keys = xrenew(const char *, keys, size);
+			keys = xrenew(sort_key_t, keys, size);
 		}
-		keys[pos++] = valid_sort_keys[i];
+		keys[pos++] = sort_key_map[i].key;
 	}
-	keys[pos] = NULL;
+	keys[pos] = SORT_INVALID;
 	return keys;
 }
 
+static const char *sort_key_to_str(sort_key_t key)
+{
+	int i;
+	for (i = 0; sort_key_map[i].str; i++) {
+		if (sort_key_map[i].key == key)
+			return sort_key_map[i].str;
+	}
+	return NULL;
+}
+
+static void sort_keys_to_str(const sort_key_t *keys, char *buf, size_t bufsize)
+{
+	int i, pos = 0;
+
+	for (i = 0; keys[i] != SORT_INVALID; i++) {
+		const char *key = sort_key_to_str(keys[i]);
+		int len = strlen(key);
+
+		if ((int)bufsize - pos - len - 2 < 0)
+			break;
+
+		memcpy(buf + pos, key, len);
+		pos += len;
+		buf[pos++] = ' ';
+	}
+	if (pos > 0)
+		pos--;
+	buf[pos] = 0;
+}
+
 static void get_lib_sort(unsigned int id, char *buf)
 {
 	strcpy(buf, lib_editable.sort_str);
@@ -250,10 +336,14 @@ static void get_lib_sort(unsigned int id, char *buf)
 
 static void set_lib_sort(unsigned int id, const char *buf)
 {
-	const char **keys = parse_sort_keys(buf);
+	sort_key_t *keys = parse_sort_keys(buf);
 
-	if (keys)
+	if (keys) {
+		editable_lock();
 		editable_set_sort_keys(&lib_editable, keys);
+		editable_unlock();
+		sort_keys_to_str(keys, lib_editable.sort_str, sizeof(lib_editable.sort_str));
+	}
 }
 
 static void get_pl_sort(unsigned int id, char *buf)
@@ -263,25 +353,32 @@ static void get_pl_sort(unsigned int id, char *buf)
 
 static void set_pl_sort(unsigned int id, const char *buf)
 {
-	const char **keys = parse_sort_keys(buf);
+	sort_key_t *keys = parse_sort_keys(buf);
 
-	if (keys)
+	if (keys) {
+		editable_lock();
 		editable_set_sort_keys(&pl_editable, keys);
+		editable_unlock();
+		sort_keys_to_str(keys, pl_editable.sort_str, sizeof(pl_editable.sort_str));
+	}
 }
 
 static void get_output_plugin(unsigned int id, char *buf)
 {
-	char *value = player_get_op();
+	const char *value = op_get_current();
 
 	if (value)
 		strcpy(buf, value);
-	free(value);
 }
 
 static void set_output_plugin(unsigned int id, const char *buf)
 {
 	if (ui_initialized) {
+		if (!soft_vol)
+			mixer_close();
 		player_set_op(buf);
+		if (!soft_vol)
+			mixer_open();
 	} else {
 		/* must set it later manually */
 		output_plugin = xstrdup(buf);
@@ -350,14 +447,7 @@ static void set_softvol_state(unsigned int id, const char *buf)
 	if (str_to_int(ptr, &r) == -1 || r < 0 || r > 100)
 		goto err;
 
-	if (soft_vol) {
-		/* avoid race condition */
-		player_set_volume(l, r);
-	} else {
-		/* can't use player_set_volume(), no race condition */
-		soft_vol_l = l;
-		soft_vol_r = r;
-	}
+	player_set_soft_volume(l, r);
 	return;
 err:
 	error_msg("two integers in range 0..100 expected");
@@ -374,7 +464,7 @@ static void set_status_display_program(unsigned int id, const char *buf)
 	free(status_display_program);
 	status_display_program = NULL;
 	if (buf[0])
-		status_display_program = xstrdup(buf);
+		status_display_program = expand_filename(buf);
 }
 
 /* }}} */
@@ -414,6 +504,24 @@ static void toggle_continue(unsigned int id)
 	update_statusline();
 }
 
+static void get_repeat_current(unsigned int id, char *buf)
+{
+	strcpy(buf, bool_names[player_repeat_current]);
+}
+
+static void set_repeat_current(unsigned int id, const char *buf)
+{
+	if (!parse_bool(buf, &player_repeat_current))
+		return;
+	update_statusline();
+}
+
+static void toggle_repeat_current(unsigned int id)
+{
+	player_repeat_current ^= 1;
+	update_statusline();
+}
+
 static void get_confirm_run(unsigned int id, char *buf)
 {
 	strcpy(buf, bool_names[confirm_run]);
@@ -483,6 +591,40 @@ static void toggle_play_sorted(unsigned int id)
 	update_statusline();
 }
 
+static void get_smart_artist_sort(unsigned int id, char *buf)
+{
+	strcpy(buf, bool_names[smart_artist_sort]);
+}
+
+static void set_smart_artist_sort(unsigned int id, const char *buf)
+{
+	if (parse_bool(buf, &smart_artist_sort))
+		tree_sort_artists();
+}
+
+static void toggle_smart_artist_sort(unsigned int id)
+{
+	smart_artist_sort ^= 1;
+	tree_sort_artists();
+}
+
+static void get_display_artist_sort_name(unsigned int id, char *buf)
+{
+	strcpy(buf, bool_names[display_artist_sort_name]);
+}
+
+static void set_display_artist_sort_name(unsigned int id, const char *buf)
+{
+	parse_bool(buf, &display_artist_sort_name);
+	lib_tree_win->changed = 1;
+}
+
+static void toggle_display_artist_sort_name(unsigned int id)
+{
+	display_artist_sort_name ^= 1;
+	lib_tree_win->changed = 1;
+}
+
 const char * const aaa_mode_names[] = {
 	"all", "artist", "album", NULL
 };
@@ -534,8 +676,8 @@ static void toggle_repeat(unsigned int id)
 	update_statusline();
 }
 
-const char * const replaygain_names[] = {
-	"disabled", "track", "album", NULL
+static const char * const replaygain_names[] = {
+	"disabled", "track", "album", "track-preferred", "album-preferred", NULL
 };
 
 static void get_replaygain(unsigned int id, char *buf)
@@ -547,14 +689,14 @@ static void set_replaygain(unsigned int id, const char *buf)
 {
 	int tmp;
 
-	if (!parse_enum(buf, 0, 2, replaygain_names, &tmp))
+	if (!parse_enum(buf, 0, 4, replaygain_names, &tmp))
 		return;
 	player_set_rg(tmp);
 }
 
 static void toggle_replaygain(unsigned int id)
 {
-	player_set_rg((replaygain + 1) % 3);
+	player_set_rg((replaygain + 1) % 5);
 }
 
 static void get_replaygain_limit(unsigned int id, char *buf)
@@ -576,6 +718,21 @@ static void toggle_replaygain_limit(unsigned int id)
 	player_set_rg_limit(replaygain_limit ^ 1);
 }
 
+static void get_resume(unsigned int id, char *buf)
+{
+	strcpy(buf, bool_names[resume_cmus]);
+}
+
+static void set_resume(unsigned int id, const char *buf)
+{
+	parse_bool(buf, &resume_cmus);
+}
+
+static void toggle_resume(unsigned int id)
+{
+	resume_cmus ^= 1;
+}
+
 static void get_show_hidden(unsigned int id, char *buf)
 {
 	strcpy(buf, bool_names[show_hidden]);
@@ -594,6 +751,23 @@ static void toggle_show_hidden(unsigned int id)
 	browser_reload();
 }
 
+static void get_show_current_bitrate(unsigned int id, char *buf)
+{
+	strcpy(buf, bool_names[show_current_bitrate]);
+}
+
+static void set_show_current_bitrate(unsigned int id, const char *buf)
+{
+	if (parse_bool(buf, &show_current_bitrate))
+		update_statusline();
+}
+
+static void toggle_show_current_bitrate(unsigned int id)
+{
+	show_current_bitrate ^= 1;
+	update_statusline();
+}
+
 static void get_show_remaining_time(unsigned int id, char *buf)
 {
 	strcpy(buf, bool_names[show_remaining_time]);
@@ -650,20 +824,46 @@ static void get_softvol(unsigned int id, char *buf)
 	strcpy(buf, bool_names[soft_vol]);
 }
 
+static void do_set_softvol(int soft)
+{
+	if (!soft_vol)
+		mixer_close();
+	player_set_soft_vol(soft);
+	if (!soft_vol)
+		mixer_open();
+	update_statusline();
+}
+
 static void set_softvol(unsigned int id, const char *buf)
 {
 	int soft;
 
 	if (!parse_bool(buf, &soft))
 		return;
-	player_set_soft_vol(soft);
+	do_set_softvol(soft);
 }
 
 static void toggle_softvol(unsigned int id)
 {
-	player_set_soft_vol(soft_vol ^ 1);
+	do_set_softvol(soft_vol ^ 1);
 }
 
+static void get_wrap_search(unsigned int id, char *buf)
+{
+	strcpy(buf, bool_names[wrap_search]);
+}
+
+static void set_wrap_search(unsigned int id, const char *buf)
+{
+	parse_bool(buf, &wrap_search);
+}
+
+static void toggle_wrap_search(unsigned int id)
+{
+	wrap_search ^= 1;
+}
+
+
 /* }}} */
 
 /* special callbacks (id set) {{{ */
@@ -714,10 +914,16 @@ static char **id_to_fmt(enum format_id id)
 		return &current_format;
 	case FMT_PLAYLIST:
 		return &list_win_format;
+	case FMT_PLAYLIST_VA:
+		return &list_win_format_va;
 	case FMT_TITLE:
 		return &window_title_format;
 	case FMT_TRACKWIN:
 		return &track_win_format;
+	case FMT_TRACKWIN_VA:
+		return &track_win_format_va;
+	default:
+		die("unhandled format code: %d\n", id);
 	}
 	return NULL;
 }
@@ -733,7 +939,7 @@ static void set_format(unsigned int id, const char *buf)
 {
 	char **fmtp = id_to_fmt(id);
 
-	if (!format_valid(buf)) {
+	if (!track_format_valid(buf)) {
 		error_msg("invalid format string");
 		return;
 	}
@@ -745,20 +951,24 @@ static void set_format(unsigned int id, const char *buf)
 
 /* }}} */
 
-#define DN(name) { #name, get_ ## name, set_ ## name, NULL },
-#define DT(name) { #name, get_ ## name, set_ ## name, toggle_ ## name },
+#define DN(name) { #name, get_ ## name, set_ ## name, NULL, 0 },
+#define DN_FLAGS(name, flags) { #name, get_ ## name, set_ ## name, NULL, flags },
+#define DT(name) { #name, get_ ## name, set_ ## name, toggle_ ## name, 0 },
 
 static const struct {
 	const char *name;
 	opt_get_cb get;
 	opt_set_cb set;
 	opt_toggle_cb toggle;
+	unsigned int flags;
 } simple_options[] = {
 	DT(aaa_mode)
 	DT(auto_reshuffle)
+	DN_FLAGS(device, OPT_PROGRAM_PATH)
 	DN(buffer_seconds)
 	DT(confirm_run)
 	DT(continue)
+	DT(smart_artist_sort)
 	DN(id3_default_charset)
 	DN(lib_sort)
 	DN(output_plugin)
@@ -766,18 +976,23 @@ static const struct {
 	DN(pl_sort)
 	DT(play_library)
 	DT(play_sorted)
+	DT(display_artist_sort_name)
 	DT(repeat)
+	DT(repeat_current)
 	DT(replaygain)
 	DT(replaygain_limit)
 	DN(replaygain_preamp)
+	DT(resume)
 	DT(show_hidden)
+	DT(show_current_bitrate)
 	DT(show_remaining_time)
 	DT(set_term_title)
 	DT(shuffle)
 	DT(softvol)
 	DN(softvol_state)
-	DN(status_display_program)
-	{ NULL, NULL, NULL, NULL }
+	DN_FLAGS(status_display_program, OPT_PROGRAM_PATH)
+	DT(wrap_search)
+	{ NULL, NULL, NULL, NULL, 0 }
 };
 
 static const char * const color_names[NR_COLORS] = {
@@ -816,12 +1031,14 @@ static const struct {
 	{ "altformat_playlist",	" %f%= %d " },
 	{ "altformat_title",	"%f" },
 	{ "altformat_trackwin",	" %f%= %d " },
-	{ "format_current",	" %a - %l - %02n. %t%= %y " },
-	{ "format_playlist",	" %-20a %02n. %t%= %y %d " },
+	{ "format_current",	" %a - %l -%3n. %t%= %y " },
+	{ "format_playlist",	" %-25%a %3n. %t%= %y %d " },
+	{ "format_playlist_va",	" %-25%A %3n. %t (%a)%= %y %d " },
 	{ "format_title",	"%a - %l - %t (%y)" },
-	{ "format_trackwin",	" %02n. %t%= %y %d " },
+	{ "format_trackwin",	"%3n. %t%= %y %d " },
+	{ "format_trackwin_va",	"%3n. %t (%a)%= %y %d " },
 
-	{ "lib_sort"	,	"artist album discnumber tracknumber title filename" },
+	{ "lib_sort"	,	"albumartist date album discnumber tracknumber title filename" },
 	{ "pl_sort",		"" },
 	{ "id3_default_charset","ISO-8859-1" },
 	{ NULL, NULL }
@@ -831,7 +1048,7 @@ LIST_HEAD(option_head);
 int nr_options = 0;
 
 void option_add(const char *name, unsigned int id, opt_get_cb get,
-		opt_set_cb set, opt_toggle_cb toggle)
+		opt_set_cb set, opt_toggle_cb toggle, unsigned int flags)
 {
 	struct cmus_opt *opt = xnew(struct cmus_opt, 1);
 	struct list_head *item;
@@ -841,6 +1058,7 @@ void option_add(const char *name, unsigned int id, opt_get_cb get,
 	opt->get = get;
 	opt->set = set;
 	opt->toggle = toggle;
+	opt->flags = flags;
 
 	item = option_head.next;
 	while (item != &option_head) {
@@ -875,51 +1093,23 @@ void option_set(const char *name, const char *value)
 		opt->set(opt->id, value);
 }
 
-static void get_op_option(unsigned int id, char *buf)
-{
-	char *val = NULL;
-
-	player_get_op_option(id, &val);
-	if (val) {
-		strcpy(buf, val);
-		free(val);
-	}
-}
-
-static void set_op_option(unsigned int id, const char *buf)
-{
-	int rc = player_set_op_option(id, buf);
-
-	if (rc) {
-		char *msg = op_get_error_msg(rc, "setting option");
-		error_msg("%s", msg);
-		free(msg);
-	}
-}
-
-/* id is ((plugin_index << 16) | option_index) */
-static void add_op_option(unsigned int id, const char *name)
-{
-	option_add(xstrdup(name), id, get_op_option, set_op_option, NULL);
-}
-
 void options_add(void)
 {
 	int i;
 
-	/* add options */
-
 	for (i = 0; simple_options[i].name; i++)
 		option_add(simple_options[i].name, 0, simple_options[i].get,
-				simple_options[i].set, simple_options[i].toggle);
+				simple_options[i].set, simple_options[i].toggle,
+				simple_options[i].flags);
 
 	for (i = 0; i < NR_FMTS; i++)
-		option_add(str_defaults[i].name, i, get_format, set_format, NULL);
+		option_add(str_defaults[i].name, i, get_format, set_format, NULL, 0);
 
 	for (i = 0; i < NR_COLORS; i++)
-		option_add(color_names[i], i, get_color, set_color, NULL);
+		option_add(color_names[i], i, get_color, set_color, NULL, 0);
 
-	player_for_each_op_option(add_op_option);
+	ip_add_options();
+	op_add_options();
 }
 
 static int handle_line(void *data, const char *line)
@@ -939,6 +1129,7 @@ void options_load(void)
 	int i;
 
 	/* initialize those that can't be statically initialized */
+	cdda_device = get_default_cdda_device();
 	for (i = 0; str_defaults[i].name; i++)
 		option_set(str_defaults[i].name, str_defaults[i].value);
 
@@ -1015,3 +1206,142 @@ void options_exit(void)
 
 	fclose(f);
 }
+
+struct resume {
+	enum player_status status;
+	char *filename;
+	long int position;
+	char *lib_filename;
+	int view;
+	char *live_filter;
+	char *browser_dir;
+};
+
+static int handle_resume_line(void *data, const char *line)
+{
+	struct resume *resume = data;
+	char *cmd, *arg;
+
+	if (!parse_command(line, &cmd, &arg))
+		return 0;
+	if (!arg)
+		goto out;
+
+	if (strcmp(cmd, "status") == 0) {
+		parse_enum(arg, 0, NR_PLAYER_STATUS, player_status_names, (int *) &resume->status);
+	} else if (strcmp(cmd, "file") == 0) {
+		free(resume->filename);
+		resume->filename = xstrdup(unescape(arg));
+	} else if (strcmp(cmd, "position") == 0) {
+		str_to_int(arg, &resume->position);
+	} else if (strcmp(cmd, "lib_file") == 0) {
+		free(resume->lib_filename);
+		resume->lib_filename = xstrdup(unescape(arg));
+	} else if (strcmp(cmd, "view") == 0) {
+		parse_enum(arg, 0, NR_VIEWS, view_names, &resume->view);
+	} else if (strcmp(cmd, "live-filter") == 0) {
+		free(resume->live_filter);
+		resume->live_filter = xstrdup(unescape(arg));
+	} else if (strcmp(cmd, "browser-dir") == 0) {
+		free(resume->browser_dir);
+		resume->browser_dir = xstrdup(unescape(arg));
+	}
+
+	free(arg);
+out:
+	free(cmd);
+	return 0;
+}
+
+void resume_load(void)
+{
+	char filename[512];
+	struct track_info *ti, *old;
+	struct resume resume = { .status = PLAYER_STATUS_STOPPED, .view = -1 };
+
+	snprintf(filename, sizeof(filename), "%s/resume", cmus_config_dir);
+	if (file_for_each_line(filename, handle_resume_line, &resume) == -1) {
+		if (errno != ENOENT)
+			error_msg("loading %s: %s", filename, strerror(errno));
+		return;
+	}
+	if (resume.view >= 0 && resume.view != cur_view)
+		set_view(resume.view);
+	if (resume.lib_filename) {
+		cache_lock();
+		ti = old = cache_get_ti(resume.lib_filename);
+		cache_unlock();
+		if (ti) {
+			editable_lock();
+			lib_add_track(ti);
+			track_info_unref(ti);
+			lib_store_cur_track(ti);
+			track_info_unref(ti);
+			ti = lib_set_track(lib_find_track(ti));
+			if (ti) {
+				BUG_ON(ti != old);
+				track_info_unref(ti);
+				tree_sel_current();
+				sorted_sel_current();
+			}
+			editable_unlock();
+		}
+		free(resume.lib_filename);
+	}
+	if (resume.filename) {
+		cache_lock();
+		ti = cache_get_ti(resume.filename);
+		cache_unlock();
+		if (ti) {
+			player_set_file(ti);
+			if (resume.status != PLAYER_STATUS_STOPPED)
+				player_seek(resume.position, 0, resume.status == PLAYER_STATUS_PLAYING);
+		}
+		free(resume.filename);
+	}
+	if (resume.live_filter) {
+		editable_lock();
+		filters_set_live(resume.live_filter);
+		editable_unlock();
+		free(resume.live_filter);
+	}
+	if (resume.browser_dir) {
+		browser_chdir(resume.browser_dir);
+		free(resume.browser_dir);
+	}
+}
+
+void resume_exit(void)
+{
+	char filename[512];
+	struct track_info *ti;
+	FILE *f;
+
+	snprintf(filename, sizeof(filename), "%s/resume", cmus_config_dir);
+	f = fopen(filename, "w");
+	if (!f) {
+		warn_errno("creating %s", filename);
+		return;
+	}
+
+	player_info_lock();
+	fprintf(f, "status %s\n", player_status_names[player_info.status]);
+	ti = player_info.ti;
+	if (ti) {
+		fprintf(f, "file %s\n", escape(ti->filename));
+		fprintf(f, "position %d\n", player_info.pos);
+	}
+	player_info_unlock();
+	if (lib_cur_track)
+		ti = tree_track_info(lib_cur_track);
+	else
+		ti = lib_get_cur_stored_track();
+	if (ti)
+		fprintf(f, "lib_file %s\n", escape(ti->filename));
+	fprintf(f, "view %s\n", view_names[cur_view]);
+	if (lib_live_filter)
+		fprintf(f, "live-filter %s\n", escape(lib_live_filter));
+	fprintf(f, "browser-dir %s\n", escape(browser_dir));
+
+	fclose(f);
+}
