@@ -1,5 +1,19 @@
 /*
+ * Copyright 2008-2011 Various Authors
  * Copyright 2005 Timo Hirvonen
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #ifndef OPTIONS_H
@@ -13,6 +27,10 @@ typedef void (*opt_get_cb)(unsigned int id, char *buf);
 typedef void (*opt_set_cb)(unsigned int id, const char *buf);
 typedef void (*opt_toggle_cb)(unsigned int id);
 
+enum {
+	OPT_PROGRAM_PATH = 1 << 0,
+};
+
 struct cmus_opt {
 	struct list_head node;
 
@@ -31,6 +49,8 @@ struct cmus_opt {
 
 	/* NULL if not toggle-able */
 	opt_toggle_cb toggle;
+
+	unsigned int flags;
 };
 
 extern struct list_head option_head;
@@ -81,17 +101,23 @@ enum {
 
 #define BRIGHT (1 << 3)
 
+extern char *cdda_device;
 extern char *output_plugin;
 extern char *status_display_program;
 extern char *server_password;
 extern int auto_reshuffle;
 extern int confirm_run;
+extern int resume_cmus;
 extern int show_hidden;
+extern int show_current_bitrate;
 extern int show_remaining_time;
 extern int set_term_title;
+extern int wrap_search;
 extern int play_library;
 extern int repeat;
 extern int shuffle;
+extern int display_artist_sort_name;
+extern int smart_artist_sort;
 
 extern const char * const aaa_mode_names[];
 extern const char * const view_names[NR_VIEWS + 1];
@@ -100,10 +126,12 @@ extern int colors[NR_COLORS];
 
 /* format string for track window (tree view) */
 extern char *track_win_format;
+extern char *track_win_format_va;
 extern char *track_win_alt_format;
 
 /* format string for shuffle, sorted and play queue views */
 extern char *list_win_format;
+extern char *list_win_format_va;
 extern char *list_win_alt_format;
 
 /* format string for currently playing track */
@@ -127,8 +155,13 @@ int source_file(const char *filename);
 /* save options */
 void options_exit(void);
 
+/* load resume file */
+void resume_load(void);
+/* save resume file */
+void resume_exit(void);
+
 void option_add(const char *name, unsigned int id, opt_get_cb get,
-		opt_set_cb set, opt_toggle_cb toggle);
+		opt_set_cb set, opt_toggle_cb toggle, unsigned int flags);
 struct cmus_opt *option_find(const char *name);
 void option_set(const char *name, const char *value);
 int parse_enum(const char *buf, int minval, int maxval, const char * const names[], int *val);
