@@ -1,4 +1,5 @@
 /*
+ * Copyright 2008-2011 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -12,9 +13,7 @@
  * General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "ui_curses.h"
@@ -27,6 +26,7 @@
 #include "filters.h"
 #include "cmus.h"
 #include "player.h"
+#include "output.h"
 #include "utils.h"
 #include "lib.h"
 #include "pl.h"
@@ -34,6 +34,7 @@
 #include "xstrjoin.h"
 #include "window.h"
 #include "format_print.h"
+#include "comment.h"
 #include "misc.h"
 #include "prog.h"
 #include "uchar.h"
@@ -42,19 +43,32 @@
 #include "keys.h"
 #include "debug.h"
 #include "help.h"
+#include "worker.h"
+#include "input.h"
+#include "file.h"
+#include "path.h"
+#include "mixer.h"
+#ifdef HAVE_CONFIG
+#include "config/curses.h"
+#include "config/iconv.h"
+#endif
 
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
 #include <sys/ioctl.h>
+#include <sys/select.h>
 #include <ctype.h>
 #include <dirent.h>
 #include <locale.h>
 #include <langinfo.h>
+#ifdef HAVE_ICONV
 #include <iconv.h>
+#endif
 #include <signal.h>
 #include <stdarg.h>
+#include <math.h>
 
 #if defined(__sun__) || defined(__CYGWIN__)
 /* TIOCGWINSZ */
@@ -76,12 +90,20 @@ enum ui_input_mode input_mode = NORMAL_MODE;
 int cur_view = TREE_VIEW;
 struct searchable *searchable;
 char *lib_filename = NULL;
+char *lib_ext_filename = NULL;
 char *pl_filename = NULL;
+char *pl_ext_filename = NULL;
+char *play_queue_filename = NULL;
+char *play_queue_ext_filename = NULL;
+char *charset = NULL;
+int using_utf8 = 0;
+
 
 /* ------------------------------------------------------------------------- */
 
 static char *lib_autosave_filename;
 static char *pl_autosave_filename;
+static char *play_queue_autosave_filename;
 
 /* shown error message and time stamp
  * error is cleared if it is older than 3s and key was pressed
@@ -94,15 +116,18 @@ static int error_count = 0;
 
 static char *server_address = NULL;
 
-static char *charset = NULL;
 static char print_buffer[512];
 
 /* destination buffer for utf8_encode and utf8_decode */
 static char conv_buffer[512];
 
-#define print_buffer_size (sizeof(print_buffer) - 1)
-static int using_utf8;
+/* one character can take up to 4 bytes in UTF-8 */
+#define print_buffer_max_width (sizeof(print_buffer) / 4 - 1)
 
+/* used for messages to the client */
+static int client_fd = -1;
+
+static char tcap_buffer[64];
 static const char *t_ts;
 static const char *t_fs;
 
@@ -114,6 +139,7 @@ static int track_win_x = 0;
 static int track_win_y = 0;
 static int track_win_w = 0;
 
+static int show_cursor;
 static int cursor_x;
 static int cursor_y;
 
@@ -189,30 +215,78 @@ static unsigned char cursed_to_fg_idx[NR_CURSED] = {
 static int pairs[NR_CURSED];
 
 enum {
+	TF_ALBUMARTIST,
 	TF_ARTIST,
 	TF_ALBUM,
 	TF_DISC,
 	TF_TRACK,
 	TF_TITLE,
 	TF_YEAR,
+	TF_ORIGINALYEAR,
 	TF_GENRE,
+	TF_COMMENT,
 	TF_DURATION,
+	TF_BITRATE,
+	TF_CODEC,
+	TF_CODEC_PROFILE,
 	TF_PATHFILE,
 	TF_FILE,
+	TF_RG_TRACK_GAIN,
+	TF_RG_TRACK_PEAK,
+	TF_RG_ALBUM_GAIN,
+	TF_RG_ALBUM_PEAK,
+	TF_ARRANGER,
+	TF_COMPOSER,
+	TF_CONDUCTOR,
+	TF_LYRICIST,
+	TF_PERFORMER,
+	TF_REMIXER,
+	TF_LABEL,
+	TF_PUBLISHER,
+	TF_WORK,
+	TF_OPUS,
+	TF_PARTNUMBER,
+	TF_PART,
+	TF_SUBTITLE,
+	TF_MEDIA,
 	NR_TFS
 };
 
 static struct format_option track_fopts[NR_TFS + 1] = {
-	DEF_FO_STR('a'),
-	DEF_FO_STR('l'),
-	DEF_FO_INT('D'),
-	DEF_FO_INT('n'),
-	DEF_FO_STR('t'),
-	DEF_FO_STR('y'),
-	DEF_FO_STR('g'),
-	DEF_FO_TIME('d'),
-	DEF_FO_STR('f'),
-	DEF_FO_STR('F'),
+	DEF_FO_STR('A', "albumartist", 0),
+	DEF_FO_STR('a', "artist", 0),
+	DEF_FO_STR('l', "album", 0),
+	DEF_FO_INT('D', "discnumber", 1),
+	DEF_FO_INT('n', "tracknumber", 1),
+	DEF_FO_STR('t', "title", 0),
+	DEF_FO_STR('y', "date", 1),
+	DEF_FO_STR('\0', "originaldate", 1),
+	DEF_FO_STR('g', "genre", 0),
+	DEF_FO_STR('c', "comment", 0),
+	DEF_FO_TIME('d', "duration", 0),
+	DEF_FO_INT('\0', "bitrate", 0),
+	DEF_FO_STR('\0', "codec", 0),
+	DEF_FO_STR('\0', "codec_profile", 0),
+	DEF_FO_STR('f', "path", 0),
+	DEF_FO_STR('F', "filename", 0),
+	DEF_FO_DOUBLE('\0', "rg_track_gain", 0),
+	DEF_FO_DOUBLE('\0', "rg_track_peak", 0),
+	DEF_FO_DOUBLE('\0', "rg_album_gain", 0),
+	DEF_FO_DOUBLE('\0', "rg_album_peak", 0),
+	DEF_FO_STR('\0', "arranger", 0),
+	DEF_FO_STR('\0', "composer", 0),
+	DEF_FO_STR('\0', "conductor", 0),
+	DEF_FO_STR('\0', "lyricist", 0),
+	DEF_FO_STR('\0', "performer", 0),
+	DEF_FO_STR('\0', "remixer", 0),
+	DEF_FO_STR('\0', "label", 0),
+	DEF_FO_STR('\0', "publisher", 0),
+	DEF_FO_STR('\0', "work", 0),
+	DEF_FO_STR('\0', "opus", 0),
+	DEF_FO_STR('\0', "partnumber", 0),
+	DEF_FO_STR('\0', "part", 0),
+	DEF_FO_STR('\0', "subtitle", 0),
+	DEF_FO_STR('\0', "media", 0),
 	DEF_FO_END
 };
 
@@ -229,27 +303,36 @@ enum {
 	SF_CONTINUE,
 	SF_SHUFFLE,
 	SF_PLAYLISTMODE,
+	SF_BITRATE,
 	NR_SFS
 };
 
 static struct format_option status_fopts[NR_SFS + 1] = {
-	DEF_FO_STR('s'),
-	DEF_FO_TIME('p'),
-	DEF_FO_TIME('d'),
-	DEF_FO_TIME('t'),
-	DEF_FO_INT('v'),
-	DEF_FO_INT('l'),
-	DEF_FO_INT('r'),
-	DEF_FO_INT('b'),
-	DEF_FO_STR('R'),
-	DEF_FO_STR('C'),
-	DEF_FO_STR('S'),
-	DEF_FO_STR('L'),
+	DEF_FO_STR('s', NULL, 0),
+	DEF_FO_TIME('p', NULL, 0),
+	DEF_FO_TIME('d', NULL, 0),
+	DEF_FO_TIME('t', NULL, 0),
+	DEF_FO_INT('v', NULL, 0),
+	DEF_FO_INT('l', NULL, 0),
+	DEF_FO_INT('r', NULL, 0),
+	DEF_FO_INT('b', NULL, 0),
+	DEF_FO_STR('R', NULL, 0),
+	DEF_FO_STR('C', NULL, 0),
+	DEF_FO_STR('S', NULL, 0),
+	DEF_FO_STR('L', NULL, 0),
+	DEF_FO_INT('B', NULL, 0),
 	DEF_FO_END
 };
 
+int track_format_valid(const char *format)
+{
+	return format_valid(format, track_fopts);
+}
+
 static void utf8_encode(const char *buffer)
 {
+	int n;
+#ifdef HAVE_ICONV
 	static iconv_t cd = (iconv_t)-1;
 	size_t is, os;
 	const char *i;
@@ -261,7 +344,7 @@ static void utf8_encode(const char *buffer)
 		cd = iconv_open("UTF-8", charset);
 		if (cd == (iconv_t)-1) {
 			d_print("iconv_open failed: %s\n", strerror(errno));
-			return;
+			goto fallback;
 		}
 	}
 	i = buffer;
@@ -272,12 +355,20 @@ static void utf8_encode(const char *buffer)
 	*o = 0;
 	if (rc == -1) {
 		d_print("iconv failed: %s\n", strerror(errno));
-		return;
+		goto fallback;
 	}
+	return;
+fallback:
+#endif
+	n = min(sizeof(conv_buffer) - 1, strlen(buffer));
+	memmove(conv_buffer, buffer, n);
+	conv_buffer[n] = '\0';
 }
 
 static void utf8_decode(const char *buffer)
 {
+	int n;
+#ifdef HAVE_ICONV
 	static iconv_t cd = (iconv_t)-1;
 	size_t is, os;
 	const char *i;
@@ -289,7 +380,7 @@ static void utf8_decode(const char *buffer)
 		cd = iconv_open(charset, "UTF-8");
 		if (cd == (iconv_t)-1) {
 			d_print("iconv_open failed: %s\n", strerror(errno));
-			return;
+			goto fallback;
 		}
 	}
 	i = buffer;
@@ -300,8 +391,13 @@ static void utf8_decode(const char *buffer)
 	*o = 0;
 	if (rc == -1) {
 		d_print("iconv failed: %s\n", strerror(errno));
-		return;
+		goto fallback;
 	}
+	return;
+fallback:
+#endif
+	n = u_to_ascii(conv_buffer, buffer, sizeof(conv_buffer) - 1);
+	conv_buffer[n] = '\0';
 }
 
 /* screen updates {{{ */
@@ -309,10 +405,10 @@ static void utf8_decode(const char *buffer)
 static void dump_print_buffer(int row, int col)
 {
 	if (using_utf8) {
-		mvaddstr(row, col, print_buffer);
+		(void) mvaddstr(row, col, print_buffer);
 	} else {
 		utf8_decode(print_buffer);
-		mvaddstr(row, col, conv_buffer);
+		(void) mvaddstr(row, col, conv_buffer);
 	}
 }
 
@@ -329,7 +425,7 @@ static int format_str(char *buf, const char *str, int width)
 		uchar u;
 		int w;
 
-		u_get_char(str, &s, &u);
+		u = u_get_char(str, &s);
 		if (u == 0) {
 			memset(buf + d, ' ', width);
 			d += width;
@@ -393,7 +489,7 @@ static void sprint_ascii(int row, int col, const char *str, int len)
 	}
 	print_buffer[len + 1] = ' ';
 	print_buffer[len + 2] = 0;
-	mvaddstr(row, col, print_buffer);
+	(void) mvaddstr(row, col, print_buffer);
 }
 
 static void print_tree(struct window *win, int row, struct iter *iter)
@@ -426,11 +522,15 @@ static void print_tree(struct window *win, int row, struct iter *iter)
 
 	pos = 0;
 	print_buffer[pos++] = ' ';
-	str = artist->name;
 	if (album) {
 		print_buffer[pos++] = ' ';
 		print_buffer[pos++] = ' ';
 		str = album->name;
+	} else {
+		if (display_artist_sort_name)
+			str = artist_sort_name(artist);
+		else
+			str = artist->name;
 	}
 	pos += format_str(print_buffer + pos, str, tree_win_w - pos - 1);
 	print_buffer[pos++] = ' ';
@@ -456,6 +556,13 @@ static inline void fopt_set_int(struct format_option *fopt, int value, int empty
 	fopt->empty = empty;
 }
 
+static inline void fopt_set_double(struct format_option *fopt, double value, int empty)
+{
+	BUG_ON(fopt->type != FO_DOUBLE);
+	fopt->fo_double = value;
+	fopt->empty = empty;
+}
+
 static inline void fopt_set_time(struct format_option *fopt, int value, int empty)
 {
 	BUG_ON(fopt->type != FO_TIME);
@@ -466,7 +573,6 @@ static inline void fopt_set_time(struct format_option *fopt, int value, int empt
 static void fill_track_fopts_track_info(struct track_info *info)
 {
 	char *filename;
-	int num, disc;
 
 	if (using_utf8) {
 		filename = info->filename;
@@ -474,37 +580,54 @@ static void fill_track_fopts_track_info(struct track_info *info)
 		utf8_encode(info->filename);
 		filename = conv_buffer;
 	}
-	disc = comments_get_int(info->comments, "discnumber");
-	num = comments_get_int(info->comments, "tracknumber");
 
-	fopt_set_str(&track_fopts[TF_ARTIST], comments_get_val(info->comments, "artist"));
-	fopt_set_str(&track_fopts[TF_ALBUM], comments_get_val(info->comments, "album"));
-	fopt_set_int(&track_fopts[TF_DISC], disc, disc == -1);
-	fopt_set_int(&track_fopts[TF_TRACK], num, num == -1);
-	fopt_set_str(&track_fopts[TF_TITLE], comments_get_val(info->comments, "title"));
-	fopt_set_str(&track_fopts[TF_YEAR], comments_get_val(info->comments, "date"));
-	fopt_set_str(&track_fopts[TF_GENRE], comments_get_val(info->comments, "genre"));
+	fopt_set_str(&track_fopts[TF_ALBUMARTIST], info->albumartist);
+	fopt_set_str(&track_fopts[TF_ARTIST], info->artist);
+	fopt_set_str(&track_fopts[TF_ALBUM], info->album);
+	fopt_set_int(&track_fopts[TF_DISC], info->discnumber, info->discnumber == -1);
+	fopt_set_int(&track_fopts[TF_TRACK], info->tracknumber, info->tracknumber == -1);
+	fopt_set_str(&track_fopts[TF_TITLE], info->title);
+	fopt_set_str(&track_fopts[TF_YEAR], keyvals_get_val(info->comments, "date"));
+	fopt_set_str(&track_fopts[TF_GENRE], info->genre);
+	fopt_set_str(&track_fopts[TF_COMMENT], info->comment);
 	fopt_set_time(&track_fopts[TF_DURATION], info->duration, info->duration == -1);
+	fopt_set_double(&track_fopts[TF_RG_TRACK_GAIN], info->rg_track_gain, isnan(info->rg_track_gain));
+	fopt_set_double(&track_fopts[TF_RG_TRACK_PEAK], info->rg_track_peak, isnan(info->rg_track_peak));
+	fopt_set_double(&track_fopts[TF_RG_ALBUM_GAIN], info->rg_album_gain, isnan(info->rg_album_gain));
+	fopt_set_double(&track_fopts[TF_RG_ALBUM_PEAK], info->rg_album_peak, isnan(info->rg_album_peak));
+	fopt_set_str(&track_fopts[TF_ORIGINALYEAR], keyvals_get_val(info->comments, "originaldate"));
+	fopt_set_int(&track_fopts[TF_BITRATE], (int) (info->bitrate / 1000. + 0.5), info->bitrate == -1);
+	fopt_set_str(&track_fopts[TF_CODEC], info->codec);
+	fopt_set_str(&track_fopts[TF_CODEC_PROFILE], info->codec_profile);
 	fopt_set_str(&track_fopts[TF_PATHFILE], filename);
-	if (is_url(info->filename)) {
+	fopt_set_str(&track_fopts[TF_ARRANGER], keyvals_get_val(info->comments, "arranger"));
+	fopt_set_str(&track_fopts[TF_COMPOSER], keyvals_get_val(info->comments, "composer"));
+	fopt_set_str(&track_fopts[TF_CONDUCTOR], keyvals_get_val(info->comments, "conductor"));
+	fopt_set_str(&track_fopts[TF_LYRICIST], keyvals_get_val(info->comments, "lyricist"));
+	fopt_set_str(&track_fopts[TF_PERFORMER], keyvals_get_val(info->comments, "performer"));
+	fopt_set_str(&track_fopts[TF_REMIXER], keyvals_get_val(info->comments, "remixer"));
+	fopt_set_str(&track_fopts[TF_LABEL], keyvals_get_val(info->comments, "label"));
+	fopt_set_str(&track_fopts[TF_PUBLISHER], keyvals_get_val(info->comments, "publisher"));
+	fopt_set_str(&track_fopts[TF_WORK], keyvals_get_val(info->comments, "work"));
+	fopt_set_str(&track_fopts[TF_OPUS], keyvals_get_val(info->comments, "opus"));
+	fopt_set_str(&track_fopts[TF_PARTNUMBER], keyvals_get_val(info->comments, "partnumber"));
+	fopt_set_str(&track_fopts[TF_PART], keyvals_get_val(info->comments, "part"));
+	fopt_set_str(&track_fopts[TF_SUBTITLE], keyvals_get_val(info->comments, "subtitle"));
+	fopt_set_str(&track_fopts[TF_MEDIA], keyvals_get_val(info->comments, "media"));
+	if (is_http_url(info->filename)) {
 		fopt_set_str(&track_fopts[TF_FILE], filename);
 	} else {
-		const char *f;
-
-		f = strrchr(filename, '/');
-		if (f) {
-			fopt_set_str(&track_fopts[TF_FILE], f + 1);
-		} else {
-			fopt_set_str(&track_fopts[TF_FILE], filename);
-		}
+		fopt_set_str(&track_fopts[TF_FILE], path_basename(filename));
 	}
 }
 
 static void print_track(struct window *win, int row, struct iter *iter)
 {
 	struct tree_track *track;
+	struct track_info *ti;
 	struct iter sel;
 	int current, selected, active;
+	const char *format;
 
 	track = iter_to_tree_track(iter);
 	current = lib_cur_track == track;
@@ -518,10 +641,16 @@ static void print_track(struct window *win, int row, struct iter *iter)
 		cursor_y = 1 + row;
 	}
 
-	fill_track_fopts_track_info(tree_track_info(track));
+	ti = tree_track_info(track);
+	fill_track_fopts_track_info(ti);
 
-	if (track_info_has_tag(tree_track_info(track))) {
-		format_print(print_buffer, track_win_w, track_win_format, track_fopts);
+	if (track_info_has_tag(ti)) {
+		if (track_is_compilation(ti->comments))
+			format = track_win_format_va;
+		else
+			format = track_win_format;
+
+		format_print(print_buffer, track_win_w, format, track_fopts);
 	} else {
 		format_print(print_buffer, track_win_w, track_win_alt_format, track_fopts);
 	}
@@ -536,6 +665,7 @@ static void print_editable(struct window *win, int row, struct iter *iter)
 	struct simple_track *track;
 	struct iter sel;
 	int current, selected, active;
+	const char *format;
 
 	track = iter_to_simple_track(iter);
 	current = current_track == track;
@@ -558,7 +688,12 @@ static void print_editable(struct window *win, int row, struct iter *iter)
 	fill_track_fopts_track_info(track->info);
 
 	if (track_info_has_tag(track->info)) {
-		format_print(print_buffer, COLS, list_win_format, track_fopts);
+		if (track_is_compilation(track->info->comments))
+			format = list_win_format_va;
+		else
+			format = list_win_format;
+
+		format_print(print_buffer, COLS, format, track_fopts);
 	} else {
 		format_print(print_buffer, COLS, list_win_alt_format, track_fopts);
 	}
@@ -613,6 +748,7 @@ static void print_filter(struct window *win, int row, struct iter *iter)
 	int current = !!e->act_stat;
 	const char stat_chars[3] = " *!";
 	int ch1, ch2, ch3, pos;
+	const char *e_filter;
 
 	window_get_sel(win, &sel);
 	selected = iters_equal(iter, &sel);
@@ -630,7 +766,14 @@ static void print_filter(struct window *win, int row, struct iter *iter)
 		ch3 = ']';
 	}
 	ch2 = stat_chars[e->sel_stat];
-	snprintf(buf, sizeof(buf), "%c%c%c%-15s  %s", ch1, ch2, ch3, e->name, e->filter);
+
+	e_filter = e->filter;
+	if (!using_utf8) {
+		utf8_encode(e_filter);
+		e_filter = conv_buffer;
+	}
+
+	snprintf(buf, sizeof(buf), "%c%c%c%-15s  %s", ch1, ch2, ch3, e->name, e_filter);
 	pos = format_str(print_buffer, buf, COLS - 1);
 	print_buffer[pos++] = ' ';
 	print_buffer[pos] = 0;
@@ -830,10 +973,10 @@ static void draw_separator(void)
 	int row;
 
 	bkgdset(pairs[CURSED_WIN_TITLE]);
-	mvaddch(0, tree_win_w, ' ');
+	(void) mvaddch(0, tree_win_w, ' ');
 	bkgdset(pairs[CURSED_SEPARATOR]);
 	for (row = 1; row < LINES - 3; row++)
-		mvaddch(row, tree_win_w, ACS_VLINE);
+		(void) mvaddch(row, tree_win_w, ACS_VLINE);
 }
 
 static void do_update_view(int full)
@@ -850,11 +993,13 @@ static void do_update_view(int full)
 			update_track_window();
 		editable_unlock();
 		draw_separator();
+		update_filterline();
 		break;
 	case SORTED_VIEW:
 		editable_lock();
 		update_sorted_window();
 		editable_unlock();
+		update_filterline();
 		break;
 	case PLAYLIST_VIEW:
 		editable_lock();
@@ -903,11 +1048,14 @@ static void do_update_statusline(void)
 	if (player_info.ti)
 		duration = player_info.ti->duration;
 
-	if (volume_max == 0) {
-		vol_left = vol_right = vol = -1;
-	} else {
-		vol_left = scale_to_percentage(player_info.vol_left, volume_max);
-		vol_right = scale_to_percentage(player_info.vol_right, volume_max);
+	vol_left = vol_right = vol = -1;
+	if (soft_vol) {
+		vol_left = soft_vol_l;
+		vol_right = soft_vol_r;
+		vol = (vol_left + vol_right + 1) / 2;
+	} else if (volume_max && volume_l >= 0 && volume_r >= 0) {
+		vol_left = scale_to_percentage(volume_l, volume_max);
+		vol_right = scale_to_percentage(volume_r, volume_max);
 		vol = (vol_left + vol_right + 1) / 2;
 	}
 	buffer_fill = scale_to_percentage(player_info.buffer_fill, player_info.buffer_size);
@@ -926,22 +1074,29 @@ static void do_update_statusline(void)
 	fopt_set_int(&status_fopts[SF_RVOLUME], vol_right, 0);
 	fopt_set_int(&status_fopts[SF_BUFFER], buffer_fill, 0);
 	fopt_set_str(&status_fopts[SF_CONTINUE], cont_strs[player_cont]);
+	fopt_set_int(&status_fopts[SF_BITRATE], player_info.current_bitrate / 1000. + 0.5, 0);
 
 	strcpy(format, " %s %p ");
 	if (duration != -1)
 		strcat(format, "/ %d ");
 	strcat(format, "- %t ");
-	if (volume_max != 0) {
-		if (player_info.vol_left != player_info.vol_right) {
+	if (vol >= 0) {
+		if (vol_left != vol_right) {
 			strcat(format, "vol: %l,%r ");
 		} else {
 			strcat(format, "vol: %v ");
 		}
 	}
-	if (player_info.ti && is_url(player_info.ti->filename))
-		strcat(format, "buf: %b ");
+	if (player_info.ti) {
+		if (is_http_url(player_info.ti->filename))
+			strcat(format, "buf: %b ");
+		if (show_current_bitrate && player_info.current_bitrate >= 0)
+			strcat(format, " %B kbps ");
+	}
 	strcat(format, "%=");
-	if (play_library) {
+	if (player_repeat_current) {
+		strcat(format, "repeat current");
+	} else if (play_library) {
 		/* artist/album modes work only in lib */
 		if (shuffle) {
 			/* shuffle overrides sorted mode */
@@ -1122,14 +1277,19 @@ static void do_update_titleline(void)
 		char *wtitle;
 
 		fill_track_fopts_track_info(player_info.ti);
-		if (is_url(player_info.ti->filename)) {
+
+		use_alt_format = !track_info_has_tag(player_info.ti);
+
+		if (is_http_url(player_info.ti->filename)) {
 			const char *title = get_stream_title();
 
-			if (title == NULL)
-				use_alt_format = 1;
-			fopt_set_str(&track_fopts[TF_TITLE], title);
-		} else {
-			use_alt_format = !track_info_has_tag(player_info.ti);
+			if (title != NULL) {
+				/*
+				 * StreamTitle overrides radio station name
+				 */
+				use_alt_format = 0;
+				fopt_set_str(&track_fopts[TF_TITLE], title);
+			}
 		}
 
 		if (use_alt_format) {
@@ -1141,10 +1301,10 @@ static void do_update_titleline(void)
 
 		/* set window title */
 		if (use_alt_format) {
-			format_print(print_buffer, sizeof(print_buffer) - 1,
+			format_print(print_buffer, print_buffer_max_width,
 					window_title_alt_format, track_fopts);
 		} else {
-			format_print(print_buffer, sizeof(print_buffer) - 1,
+			format_print(print_buffer,  print_buffer_max_width,
 					window_title_format, track_fopts);
 		}
 
@@ -1226,7 +1386,13 @@ static void post_update(void)
 			move(LINES - 1, 0);
 		}
 		refresh();
-		curs_set(0);
+
+		/* visible cursor is useful for screen readers */
+		if (show_cursor) {
+			curs_set(1);
+		} else {
+			curs_set(0);
+		}
 	}
 }
 
@@ -1269,6 +1435,20 @@ void update_statusline(void)
 	post_update();
 }
 
+void update_filterline(void)
+{
+	if (cur_view != TREE_VIEW && cur_view != SORTED_VIEW)
+		return;
+	if (lib_live_filter) {
+		char buf[512];
+		int w;
+		bkgdset(pairs[CURSED_STATUSLINE]);
+		snprintf(buf, sizeof(buf), "filtered: %s", lib_live_filter);
+		w = clamp(strlen(buf) + 2, COLS/4, COLS/2);
+		sprint(LINES-4, COLS-w, buf, w);
+	}
+}
+
 void info_msg(const char *format, ...)
 {
 	va_list ap;
@@ -1277,6 +1457,11 @@ void info_msg(const char *format, ...)
 	vsnprintf(error_buf, sizeof(error_buf), format, ap);
 	va_end(ap);
 
+	if (client_fd != -1) {
+		write_all(client_fd, error_buf, strlen(error_buf));
+		write_all(client_fd, "\n", 1);
+	}
+
 	msg_is_error = 0;
 
 	update_commandline();
@@ -1292,6 +1477,10 @@ void error_msg(const char *format, ...)
 	va_end(ap);
 
 	d_print("%s\n", error_buf);
+	if (client_fd != -1) {
+		write_all(client_fd, error_buf, strlen(error_buf));
+		write_all(client_fd, "\n", 1);
+	}
 
 	msg_is_error = 1;
 	error_count++;
@@ -1383,7 +1572,17 @@ void search_not_found(void)
 			break;
 		}
 	}
-	info_msg("%s not found: %s", what, search_str ? : "");
+	info_msg("%s not found: %s", what, search_str ? search_str : "");
+}
+
+void set_client_fd(int fd)
+{
+	client_fd = fd;
+}
+
+int get_client_fd(void)
+{
+	return client_fd;
 }
 
 void set_view(int view)
@@ -1490,61 +1689,71 @@ static void clear_error(void)
 
 static void spawn_status_program(void)
 {
-	static const char *status_strs[] = { "stopped", "playing", "paused" };
+	enum player_status status;
 	const char *stream_title = NULL;
 	char *argv[32];
-	int i, status;
+	int i;
 
 	if (status_display_program == NULL || status_display_program[0] == 0)
 		return;
 
 	player_info_lock();
 	status = player_info.status;
-	if (status == 1)
+	if (status == PLAYER_STATUS_PLAYING && player_info.ti && is_http_url(player_info.ti->filename))
 		stream_title = get_stream_title();
 
 	i = 0;
 	argv[i++] = xstrdup(status_display_program);
 
 	argv[i++] = xstrdup("status");
-	argv[i++] = xstrdup(status_strs[status]);
+	argv[i++] = xstrdup(player_status_names[status]);
 	if (player_info.ti) {
 		static const char *keys[] = {
-			"artist", "album", "discnumber", "tracknumber", "title", "date", NULL
+			"artist", "album", "discnumber", "tracknumber", "title", "date",
+			"musicbrainz_trackid", NULL
 		};
 		int j;
 
-		if (is_url(player_info.ti->filename)) {
+		if (is_http_url(player_info.ti->filename)) {
 			argv[i++] = xstrdup("url");
-			argv[i++] = xstrdup(player_info.ti->filename);
-			if (stream_title) {
-				argv[i++] = xstrdup("title");
-				argv[i++] = xstrdup(stream_title);
-			}
 		} else {
-			char buf[32];
-
 			argv[i++] = xstrdup("file");
-			argv[i++] = xstrdup(player_info.ti->filename);
+		}
+		argv[i++] = xstrdup(player_info.ti->filename);
+
+		if (track_info_has_tag(player_info.ti)) {
 			for (j = 0; keys[j]; j++) {
 				const char *key = keys[j];
 				const char *val;
 
-				val = comments_get_val(player_info.ti->comments, key);
+				if (strcmp(key, "title") == 0 && stream_title)
+					/*
+					 * StreamTitle overrides radio station name
+					 */
+					val = stream_title;
+				else
+					val = keyvals_get_val(player_info.ti->comments, key);
+
 				if (val) {
 					argv[i++] = xstrdup(key);
 					argv[i++] = xstrdup(val);
 				}
 			}
-			snprintf(buf, sizeof(buf), "%d", player_info.ti->duration);
-			argv[i++] = xstrdup("duration");
-			argv[i++] = xstrdup(buf);
+			if (player_info.ti->duration > 0) {
+				char buf[32];
+				snprintf(buf, sizeof(buf), "%d", player_info.ti->duration);
+				argv[i++] = xstrdup("duration");
+				argv[i++] = xstrdup(buf);
+			}
+		} else if (stream_title) {
+			argv[i++] = xstrdup("title");
+			argv[i++] = xstrdup(stream_title);
 		}
 	}
 	argv[i++] = NULL;
 	player_info_unlock();
 
-	if (spawn(argv, &status) == -1)
+	if (spawn(argv, NULL, 0) == -1)
 		error_msg("couldn't run `%s': %s", status_display_program, strerror(errno));
 	for (i = 0; argv[i]; i++)
 		free(argv[i]);
@@ -1557,6 +1766,11 @@ static void sig_int(int sig)
 	ctrl_c_pressed = 1;
 }
 
+static void sig_hup(int sig)
+{
+	cmus_running = 0;
+}
+
 static int needs_to_resize = 1;
 
 static void sig_winch(int sig)
@@ -1607,7 +1821,9 @@ static void update(void)
 
 		if (get_window_size(&lines, &columns) == 0) {
 			needs_to_resize = 0;
+#if HAVE_RESIZETERM
 			resizeterm(lines, columns);
+#endif
 			w = COLS;
 			h = LINES - 3;
 			if (w < 16)
@@ -1627,6 +1843,8 @@ static void update(void)
 			needs_status_update = 1;
 			needs_command_update = 1;
 		}
+		clearok(curscr, TRUE);
+		refresh();
 	}
 
 	player_info_lock();
@@ -1644,10 +1862,9 @@ static void update(void)
 		player_info.metadata_changed = 0;
 		needs_title_update = 1;
 	}
-	if (player_info.position_changed || player_info.status_changed || player_info.vol_changed) {
+	if (player_info.position_changed || player_info.status_changed) {
 		player_info.position_changed = 0;
 		player_info.status_changed = 0;
-		player_info.vol_changed = 0;
 
 		needs_status_update = 1;
 	}
@@ -1719,6 +1936,20 @@ static void handle_ch(uchar ch)
 	}
 }
 
+static void handle_escape(int c)
+{
+	clear_error();
+	if (input_mode == NORMAL_MODE) {
+		normal_mode_ch(c + 128);
+	} else if (input_mode == COMMAND_MODE) {
+		command_mode_escape(c);
+		update_commandline();
+	} else if (input_mode == SEARCH_MODE) {
+		search_mode_escape(c);
+		update_commandline();
+	}
+}
+
 static void handle_key(int key)
 {
 	clear_error();
@@ -1749,6 +1980,15 @@ static void u_getch(void)
 		return;
 	}
 
+	/* escape sequence */
+	if (key == 0x1B) {
+		int e_key = getch();
+		if (e_key != ERR && e_key != 0) {
+			handle_escape(e_key);
+			return;
+		}
+	}
+
 	ch = (unsigned char)key;
 	while (bit > 0 && ch & mask) {
 		mask >>= 1;
@@ -1757,7 +1997,7 @@ static void u_getch(void)
 	if (bit == 7) {
 		/* ascii */
 		u = ch;
-	} else {
+	} else if (using_utf8) {
 		int count;
 
 		u = ch & ((1 << bit) - 1);
@@ -1771,7 +2011,8 @@ static void u_getch(void)
 			u = (u << 6) | (ch & 63);
 			count--;
 		}
-	}
+	} else
+		u = ch | U_INVALID_MASK;
 	handle_ch(u);
 }
 
@@ -1783,13 +2024,14 @@ static void main_loop(void)
 	while (cmus_running) {
 		fd_set set;
 		struct timeval tv;
+		int poll_mixer = 0;
+		int i, nr_fds = 0;
+		int fds[NR_MIXER_FDS];
+		struct list_head *item;
+		struct client *client;
 
 		update();
 
-		FD_ZERO(&set);
-		FD_SET(0, &set);
-		FD_SET(server_socket, &set);
-
 		/* Timeout must be so small that screen updates seem instant.
 		 * Only affects changes done in other threads (worker, player).
 		 *
@@ -1799,37 +2041,88 @@ static void main_loop(void)
 		 * The timeout is accuracy of player position.
 		 */
 		tv.tv_sec = 0;
-		tv.tv_usec = 100e3;
-		rc = select(fd_high + 1, &set, NULL, NULL, &tv);
+		tv.tv_usec = 0;
+
+		player_info_lock();
+		if (player_info.status == PLAYER_STATUS_PLAYING) {
+			// player position updates need to be fast
+			tv.tv_usec = 100e3;
+		}
+		player_info_unlock();
+
+		if (!tv.tv_usec && worker_has_job(JOB_TYPE_ANY)) {
+			// playlist is loading. screen needs to be updated
+			tv.tv_usec = 250e3;
+		}
+
+		FD_ZERO(&set);
+		FD_SET(0, &set);
+		FD_SET(server_socket, &set);
+		list_for_each_entry(client, &client_head, node) {
+			FD_SET(client->fd, &set);
+			if (client->fd > fd_high)
+				fd_high = client->fd;
+		}
+		if (!soft_vol) {
+			nr_fds = mixer_get_fds(fds);
+			if (nr_fds <= 0) {
+				poll_mixer = 1;
+				if (!tv.tv_usec)
+					tv.tv_usec = 500e3;
+			}
+			for (i = 0; i < nr_fds; i++) {
+				BUG_ON(fds[i] <= 0);
+				FD_SET(fds[i], &set);
+				if (fds[i] > fd_high)
+					fd_high = fds[i];
+			}
+		}
+
+		if (tv.tv_usec) {
+			rc = select(fd_high + 1, &set, NULL, NULL, &tv);
+		} else {
+			rc = select(fd_high + 1, &set, NULL, NULL, NULL);
+		}
+		if (poll_mixer) {
+			int ol = volume_l;
+			int or = volume_r;
+
+			mixer_read_volume();
+			if (ol != volume_l || or != volume_r)
+				update_statusline();
+
+		}
 		if (rc <= 0) {
 			if (ctrl_c_pressed) {
 				handle_ch(0x03);
 				ctrl_c_pressed = 0;
 			}
+
 			continue;
 		}
 
-		if (FD_ISSET(server_socket, &set))
-			server_serve();
-		if (FD_ISSET(0, &set)) {
-			if (using_utf8) {
-				u_getch();
-			} else {
-				int key = getch();
-
-				if (key != ERR && key != 0) {
-					if (key > 255) {
-						handle_key(key);
-					} else {
-						uchar ch = key;
-
-						if (ch > 0x7f)
-							ch |= U_INVALID_MASK;
-						handle_ch(ch);
-					}
-				}
+		for (i = 0; i < nr_fds; i++) {
+			if (FD_ISSET(fds[i], &set)) {
+				d_print("vol changed\n");
+				mixer_read_volume();
+				update_statusline();
 			}
 		}
+		if (FD_ISSET(server_socket, &set))
+			server_accept();
+
+		// server_serve() can remove client from the list
+		item = client_head.next;
+		while (item != &client_head) {
+			struct list_head *next = item->next;
+			client = container_of(item, struct client, node);
+			if (FD_ISSET(client->fd, &set))
+				server_serve(client);
+			item = next;
+		}
+
+		if (FD_ISSET(0, &set))
+			u_getch();
 	}
 }
 
@@ -1869,6 +2162,16 @@ static void init_curses(void)
 	act.sa_handler = sig_int;
 	sigaction(SIGINT, &act, NULL);
 
+	sigemptyset(&act.sa_mask);
+	act.sa_flags = 0;
+	act.sa_handler = sig_hup;
+	sigaction(SIGHUP, &act, NULL);
+
+	sigemptyset(&act.sa_mask);
+	act.sa_flags = 0;
+	act.sa_handler = SIG_IGN;
+	sigaction(SIGPIPE, &act, NULL);
+
 	sigemptyset(&act.sa_mask);
 	act.sa_flags = 0;
 	act.sa_handler = sig_winch;
@@ -1888,8 +2191,10 @@ static void init_curses(void)
 
 	noecho();
 	if (has_colors()) {
+#if HAVE_USE_DEFAULT_COLORS
 		start_color();
 		use_default_colors();
+#endif
 	}
 	d_print("Number of supported colors: %d\n", COLORS);
 	ui_initialized = 1;
@@ -1899,7 +2204,7 @@ static void init_curses(void)
 	 */
 	update_colors();
 
-	ptr = print_buffer;
+	ptr = tcap_buffer;
 	t_ts = tgetstr("ts", &ptr);
 	t_fs = tgetstr("fs", &ptr);
 	d_print("ts: %d fs: %d\n", !!t_ts, !!t_fs);
@@ -1907,7 +2212,8 @@ static void init_curses(void)
 	if (!t_fs)
 		t_ts = NULL;
 
-	if (!t_ts && (term = getenv("TERM"))) {
+	term = getenv("TERM");
+	if (!t_ts && term) {
 		/*
 		 * Eterm:            Eterm
 		 * aterm:            rxvt
@@ -1958,40 +2264,54 @@ static void init_all(void)
 
 	/* finally we can set the output plugin */
 	player_set_op(output_plugin);
-
-	player_get_volume(&player_info.vol_left, &player_info.vol_right);
+	if (!soft_vol)
+		mixer_open();
 
 	lib_autosave_filename = xstrjoin(cmus_config_dir, "/lib.pl");
 	pl_autosave_filename = xstrjoin(cmus_config_dir, "/playlist.pl");
+	play_queue_autosave_filename = xstrjoin(cmus_config_dir, "/queue.pl");
 	pl_filename = xstrdup(pl_autosave_filename);
 	lib_filename = xstrdup(lib_autosave_filename);
 
-	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB);
-	cmus_add(pl_add_track, pl_autosave_filename, FILE_TYPE_PL, JOB_TYPE_PL);
-
 	if (error_count) {
 		char buf[16];
+		char *ret;
 
 		warn("Press <enter> to continue.");
-		fgets(buf, sizeof(buf), stdin);
+
+		ret = fgets(buf, sizeof(buf), stdin);
+		BUG_ON(ret == NULL);
 	}
 	help_add_all_unbound();
 
 	init_curses();
+
+	if (resume_cmus) {
+		resume_load();
+		cmus_add(play_queue_append, play_queue_autosave_filename, FILE_TYPE_PL, JOB_TYPE_QUEUE);
+	}
+
+	cmus_add(pl_add_track, pl_autosave_filename, FILE_TYPE_PL, JOB_TYPE_PL);
+	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB);
 }
 
 static void exit_all(void)
 {
 	endwin();
 
+	if (resume_cmus)
+		resume_exit();
 	options_exit();
 
 	server_exit();
 	cmus_exit();
+	if (resume_cmus)
+		cmus_save(play_queue_for_each, play_queue_autosave_filename);
 	cmus_save(lib_for_each, lib_autosave_filename);
 	cmus_save(pl_for_each, pl_autosave_filename);
 
 	player_exit();
+	op_exit_plugins();
 	commands_exit();
 	search_mode_exit();
 	filters_exit();
@@ -2002,6 +2322,7 @@ static void exit_all(void)
 enum {
 	FLAG_LISTEN,
 	FLAG_PLUGINS,
+	FLAG_SHOW_CURSOR,
 	FLAG_HELP,
 	FLAG_VERSION,
 	NR_FLAGS
@@ -2010,6 +2331,7 @@ enum {
 static struct option options[NR_FLAGS + 1] = {
 	{ 0, "listen", 1 },
 	{ 0, "plugins", 0 },
+	{ 0, "show-cursor", 0 },
 	{ 0, "help", 0 },
 	{ 0, "version", 0 },
 	{ 0, NULL, 0 }
@@ -2023,6 +2345,7 @@ static const char *usage =
 "                      ADDR is either a UNIX socket or host[:port]\n"
 "                      WARNING: using TCP/IP is insecure!\n"
 "      --plugins       list available plugins and exit\n"
+"      --show-cursor   always visible cursor\n"
 "      --help          display this help and exit\n"
 "      --version       " VERSION "\n"
 "\n"
@@ -2048,7 +2371,9 @@ int main(int argc, char *argv[])
 			printf(usage, program_name);
 			return 0;
 		case FLAG_VERSION:
-			printf("cmus " VERSION "\nCopyright 2004-2006 Timo Hirvonen\n");
+			printf("cmus " VERSION
+			       "\nCopyright 2004-2006 Timo Hirvonen"
+			       "\nCopyright 2008-2011 Various Authors\n");
 			return 0;
 		case FLAG_PLUGINS:
 			list_plugins = 1;
@@ -2056,29 +2381,36 @@ int main(int argc, char *argv[])
 		case FLAG_LISTEN:
 			server_address = xstrdup(arg);
 			break;
+		case FLAG_SHOW_CURSOR:
+			show_cursor = 1;
+			break;
 		}
 	}
 
 	setlocale(LC_CTYPE, "");
+	setlocale(LC_COLLATE, "");
+	charset = getenv("CMUS_CHARSET");
+	if (!charset || !charset[0]) {
 #ifdef CODESET
-	charset = nl_langinfo(CODESET);
+		charset = nl_langinfo(CODESET);
 #else
-	charset = "ISO-8859-1";
+		charset = "ISO-8859-1";
 #endif
-	if (strcmp(charset, "UTF-8") == 0) {
-		using_utf8 = 1;
-	} else {
-		using_utf8 = 0;
 	}
+	if (strcmp(charset, "UTF-8") == 0)
+		using_utf8 = 1;
+
 	misc_init();
 	if (server_address == NULL)
 		server_address = xstrjoin(cmus_config_dir, "/socket");
 	debug_init();
 	d_print("charset = '%s'\n", charset);
 
-	player_load_plugins();
+	ip_load_plugins();
+	op_load_plugins();
 	if (list_plugins) {
-		player_dump_plugins();
+		ip_dump_plugins();
+		op_dump_plugins();
 		return 0;
 	}
 	init_all();
